# -*- coding: utf-8 -*-
"""This file contains a class to provide a parsing framework to plaso.

This class contains a base framework class for parsing fileobjects, and
also some implementations that extend it to provide a more comprehensive
parser.
"""

import abc
import csv
import logging

import pyparsing

from dfvfs.helpers import text_file

from plaso.containers import events
from plaso.lib import errors
from plaso.lib import utils
from plaso.parsers import interface


# Pylint complains about some functions not being implemented that shouldn't
# be since they need to be implemented by children.
# pylint: disable=abstract-method

class TextCSVParser(interface.FileObjectParser):
  """An implementation of a simple CSV line-per-entry log files.

  Attributes:
    encoding (str): encoding used in the CSV file, or None if the encoding is
        unknown.
  """

  # A list that contains the names of all the fields in the log file.
  COLUMNS = []

  # A CSV file is comma separated, but this can be overwritten to include
  # tab, pipe or other character separation. Note this must be a byte string
  # otherwise TypeError: "delimiter" must be an 1-character string is raised.
  VALUE_SEPARATOR = b','

  # If there is a header before the lines start it can be defined here, and
  # the number of header lines that need to be skipped before the parsing
  # starts.
  NUMBER_OF_HEADER_LINES = 0

  # If there is a special quote character used inside the structured text
  # it can be defined here.
  QUOTE_CHAR = b'"'

  # Value that should not appear inside the file, made to test the actual
  # file to see if it confirms to standards.
  MAGIC_TEST_STRING = b'RegnThvotturMeistarans'

  def __init__(self, encoding=None):
    """Initializes a parser object.

    Args:
      encoding (Optional[str]): encoding used in the CSV file, where None
          indicates the codepage of the parser mediator should be used.
    """
    super(TextCSVParser, self).__init__()
    self.encoding = encoding

  def _ConvertRowToUnicode(self, parser_mediator, row):
    """Converts all strings in a CSV row dict to Unicode.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row (dict[str, bytes]): a row from a CSV file, where the dictionary
          key contains the column name and the value a binary string.

    Returns:
      dict[str, str]: a row from the CSV file, where the dictionary key
          contains the column name and the value a Unicode string.
    """
    if not self.encoding:
      # If no encoding is set, we default to the system codepage.
      self.encoding = parser_mediator.codepage

    for key, value in iter(row.items()):
      try:
        row[key] = value.decode(self.encoding)
      except UnicodeDecodeError:
        replaced_value = value.decode(self.encoding, errors=u'replace')
        parser_mediator.ProduceExtractionError(
            u'error decoding string as {0:s}, characters have been '
            u'replaced in {1:s}'.format(self.encoding, replaced_value))
        row[key] = replaced_value
    return row

  def ParseRow(self, parser_mediator, row_offset, row):
    """Parse a line of the log file and extract event objects.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row_offset: the offset of the row.
      row: a dictionary containing all the fields as denoted in the
           COLUMNS class list. Strings in this dict are Unicode strings.
    """
    # TODO: replace EventObject with a CSV specific event object.
    event_object = events.EventObject()
    if row_offset is not None:
      event_object.offset = row_offset
    event_object.row_dict = row
    parser_mediator.ProduceEvent(event_object)

  def ParseFileObject(self, parser_mediator, file_object, **unused_kwargs):
    """Parses a CSV text file-like object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      file_object (dfvfs.FileIO): file-like object.

    Raises:
      UnableToParseFile: when the file cannot be parsed.
    """
    file_entry = parser_mediator.GetFileEntry()
    path_spec_printable = file_entry.path_spec.comparable.replace(u'\n', u';')

    text_file_object = text_file.TextFile(file_object)

    # If we specifically define a number of lines we should skip, do that here.
    for _ in range(0, self.NUMBER_OF_HEADER_LINES):
      _ = text_file_object.readline()

    reader = csv.DictReader(
        text_file_object, fieldnames=self.COLUMNS,
        restkey=self.MAGIC_TEST_STRING, restval=self.MAGIC_TEST_STRING,
        delimiter=self.VALUE_SEPARATOR, quotechar=self.QUOTE_CHAR)

    try:
      row = reader.next()
    except (csv.Error, StopIteration):
      raise errors.UnableToParseFile(
          u'[{0:s}] Unable to parse CSV file: {1:s}.'.format(
              self.NAME, path_spec_printable))

    number_of_columns = len(self.COLUMNS)
    number_of_records = len(row)

    if number_of_records != number_of_columns:
      raise errors.UnableToParseFile((
          u'[{0:s}] Unable to parse CSV file: {1:s}. Wrong number of '
          u'records (expected: {2:d}, got: {3:d})').format(
              self.NAME, path_spec_printable, number_of_columns,
              number_of_records))

    for key, value in row.items():
      if key == self.MAGIC_TEST_STRING or value == self.MAGIC_TEST_STRING:
        raise errors.UnableToParseFile((
            u'[{0:s}] Unable to parse CSV file: {1:s}. Signature '
            u'mismatch.').format(self.NAME, path_spec_printable))

    if not self.VerifyRow(parser_mediator, row):
      raise errors.UnableToParseFile((
          u'[{0:s}] Unable to parse CSV file: {1:s}. Verification '
          u'failed.').format(self.NAME, path_spec_printable))

    row = self._ConvertRowToUnicode(parser_mediator, row)
    self.ParseRow(parser_mediator, text_file_object.tell(), row)

    for row in reader:
      if parser_mediator.abort:
        break
      row = self._ConvertRowToUnicode(parser_mediator, row)
      self.ParseRow(parser_mediator, text_file_object.tell(), row)

  def VerifyRow(self, unused_parser_mediator, unused_row):
    """Return a bool indicating whether or not this is the correct parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row (dict[str, bytes]): a single row from the CSV file.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """
    pass


# TODO: determine if this method should be merged with PyParseIntCast.
def ConvertTokenToInteger(unused_string, unused_location, tokens):
  """Pyparsing parse action callback to convert a token into an integer value.

  Args:
    string (str): original string.
    location (int): location in the string where the token was found.
    tokens (list[str]): tokens.

  Returns:
    int: integer value or None.
  """
  try:
    return int(tokens[0], 10)
  except ValueError:
    pass


def PyParseRangeCheck(lower_bound, upper_bound):
  """Verify that a number is within a defined range.

  This is a callback method for pyparsing setParseAction
  that verifies that a read number is within a certain range.

  To use this method it needs to be defined as a callback method
  in setParseAction with the upper and lower bound set as parameters.

  Args:
    lower_bound (int): lower bound of the range.
    upper_bound (int): upper bound of the range.

  Returns:
    Function: callback method that can be used by pyparsing setParseAction.
  """
  def CheckRange(unused_string, unused_location, tokens):
    """Parse the arguments.

    Args:
      location (int): location within the string where the match was made.
      tokens (list[str]): tokens.
    """
    try:
      check_number = tokens[0]
    except IndexError:
      check_number = -1

    if check_number < lower_bound:
      raise pyparsing.ParseException(
          u'Value: {0:d} precedes lower bound: {1:d}'.format(
              check_number, lower_bound))

    if check_number > upper_bound:
      raise pyparsing.ParseException(
          u'Value: {0:d} exceeds upper bound: {1:d}'.format(
              check_number, upper_bound))

  # Since callback methods for pyparsing need to accept certain parameters
  # and there is no way to define conditions, like upper and lower bounds
  # we need to return here a method that accepts those pyparsing parameters.
  return CheckRange


def PyParseIntCast(unused_string, unused_location, tokens):
  """Return an integer from a string.

  This is a pyparsing callback method that converts the matched
  string into an integer.

  The method modifies the content of the tokens list and converts
  them all to an integer value.

  Args:
    string (str): original parsed string.
    location (int): location within the string where the match was made.
    tokens (list[str]): extracted tokens, where the string to be converted
        is stored.
  """
  # Cast the regular tokens.
  for index, token in enumerate(tokens):
    try:
      tokens[index] = int(token)
    except ValueError:
      logging.error(u'Unable to cast [{0:s}] to an int, setting to 0'.format(
          token))
      tokens[index] = 0

  # We also need to cast the dictionary built tokens.
  for key in tokens.keys():
    try:
      tokens[key] = int(tokens[key], 10)
    except ValueError:
      logging.error(
          u'Unable to cast [{0:s} = {1:d}] to an int, setting to 0'.format(
              key, tokens[key]))
      tokens[key] = 0


def PyParseJoinList(unused_string, unused_location, tokens):
  """Return a joined token from a list of tokens.

  This is a callback method for pyparsing setParseAction that modifies
  the returned token list to join all the elements in the list to a single
  token.

  Args:
    string (str): original parsed string.
    location (int): location within the string where the match was made.
    tokens (list[str]): extracted tokens, where the string to be converted
        is stored.
  """
  join_list = []
  for token in tokens:
    try:
      join_list.append(str(token))
    except UnicodeDecodeError:
      join_list.append(repr(token))

  tokens[0] = u''.join(join_list)
  del tokens[1:]


class PyparsingConstants(object):
  """A class that maintains constants for pyparsing."""

  # Numbers.
  INTEGER = pyparsing.Word(pyparsing.nums).setParseAction(PyParseIntCast)
  IPV4_OCTET = pyparsing.Word(pyparsing.nums, min=1, max=3).setParseAction(
      PyParseIntCast, PyParseRangeCheck(0, 255))
  IPV4_ADDRESS = (IPV4_OCTET + ('.' + IPV4_OCTET) * 3).setParseAction(
      PyParseJoinList)

  # TODO: Fix the IPv6 address specification to be more accurate (8 :, correct
  # size, etc).
  IPV6_ADDRESS = pyparsing.Word(':' + pyparsing.hexnums).setParseAction(
      PyParseJoinList)

  IP_ADDRESS = (IPV4_ADDRESS | IPV6_ADDRESS)

  # TODO: deprecate and remove, use THREE_LETTERS instead.
  MONTH = pyparsing.Word(
      pyparsing.string.uppercase, pyparsing.string.lowercase, exact=3)

  # Define date structures.
  HYPHEN = pyparsing.Literal(u'-').suppress()

  ONE_OR_TWO_DIGITS = pyparsing.Word(
      pyparsing.nums, min=1, max=2).setParseAction(PyParseIntCast)
  TWO_DIGITS = pyparsing.Word(pyparsing.nums, exact=2).setParseAction(
      PyParseIntCast)
  THREE_DIGITS = pyparsing.Word(pyparsing.nums, exact=3).setParseAction(
      PyParseIntCast)
  FOUR_DIGITS = pyparsing.Word(pyparsing.nums, exact=4).setParseAction(
      PyParseIntCast)

  THREE_LETTERS = pyparsing.Word(pyparsing.alphas, exact=3)

  DATE_ELEMENTS = (
      FOUR_DIGITS.setResultsName(u'year') + pyparsing.Suppress(u'-') +
      TWO_DIGITS.setResultsName(u'month') + pyparsing.Suppress(u'-') +
      TWO_DIGITS.setResultsName(u'day_of_month'))
  TIME_ELEMENTS = (
      TWO_DIGITS.setResultsName(u'hours') + pyparsing.Suppress(':') +
      TWO_DIGITS.setResultsName(u'minutes') + pyparsing.Suppress(':') +
      TWO_DIGITS.setResultsName(u'seconds'))
  TIME_MSEC_ELEMENTS = (
      TIME_ELEMENTS + pyparsing.Suppress('.') +
      INTEGER.setResultsName(u'microseconds'))

  # Date structures defined as a single group.
  DATE = pyparsing.Group(DATE_ELEMENTS)
  DATE_TIME = pyparsing.Group(DATE_ELEMENTS + TIME_ELEMENTS)
  DATE_TIME_MSEC = pyparsing.Group(DATE_ELEMENTS + TIME_MSEC_ELEMENTS)
  TIME = pyparsing.Group(TIME_ELEMENTS)

  TIME_MSEC = TIME + pyparsing.Suppress('.') + INTEGER
  # TODO: replace by
  # TIME_MSEC = pyparsing.Group(TIME_MSEC_ELEMENTS)

  COMMENT_LINE_HASH = pyparsing.Literal(u'#') + pyparsing.SkipTo(
      pyparsing.LineEnd())
  # TODO: Add more commonly used structs that can be used by parsers.
  PID = pyparsing.Word(
      pyparsing.nums, min=1, max=5).setParseAction(PyParseIntCast)


class PyparsingSingleLineTextParser(interface.FileObjectParser):
  """Single line text parser based on the pyparsing library."""

  # The actual structure, this needs to be defined by each parser.
  # This is defined as a list of tuples so that more then a single line
  # structure can be defined. That way the parser can support more than a
  # single type of log entry, despite them all having in common the constraint
  # that each log entry is a single line.
  # The tuple should have two entries, a key and a structure. This is done to
  # keep the structures in an order of priority/preference.
  # The key is a comment or an identification that is passed to the ParseRecord
  # function so that the developer can identify which structure got parsed.
  # The value is the actual pyparsing structure.
  LINE_STRUCTURES = []

  # In order for the tool to not read too much data into a buffer to evaluate
  # whether or not the parser is the right one for this file or not we
  # specifically define a maximum amount of bytes a single line can occupy. This
  # constant can be overwritten by implementations if their format might have a
  # longer line than 400 bytes.
  MAX_LINE_LENGTH = 400

  # Define an encoding. If a file is encoded using specific encoding it is
  # advised to include it here. If this class constant is set all lines wil be
  # decoded prior to being sent to parsing by pyparsing, if not properly set it
  # could negatively affect parsing of the file.
  # If this value needs to be calculated on the fly (not a fixed constant for
  # this particular file type) it can be done by modifying the self.encoding
  # attribute.
  _ENCODING = u'ascii'

  _EMPTY_LINES = frozenset([b'\n', b'\r', b'\r\n'])

  def __init__(self):
    """Initializes a parser object."""
    super(PyparsingSingleLineTextParser, self).__init__()
    self._current_offset = 0
    # TODO: self._line_structures is a work-around and this needs
    # a structural fix.
    self._line_structures = self.LINE_STRUCTURES
    self.encoding = self._ENCODING

  def _ReadLine(
      self, parser_mediator, text_file_object, max_len=0, quiet=False, depth=0):
    """Reads a line from a text file.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      text_file_object (dfvfs.TextFile): text file.
      max_len (Optional[int]): maximum number of bytes a single line can take.
      quiet (Optional[bool]): True if parse warnings should not be displayed.
      depth (Optional[int]): number of new lines the parser can encounter
          before bailing out.

    Returns:
      str: single line read from the file-like object, or the maximum number of
          characters, if max_len defined and line longer than the defined size.
    """
    if max_len:
      line = text_file_object.readline(max_len)
    else:
      line = text_file_object.readline()

    if not line:
      return

    if line in self._EMPTY_LINES:
      # Max 40 new lines in a row before we bail out.
      if depth == 40:
        return u''

      return self._ReadLine(
          parser_mediator, text_file_object, max_len=max_len, depth=depth + 1)

    if not self.encoding:
      return line.strip()

    try:
      line = line.decode(self.encoding)
    except UnicodeDecodeError:
      if not quiet:
        parser_mediator.ProduceExtractionError(
            u'unable to decode line: "{0:s}..." with encoding: {1:s}'.format(
                repr(line[:30]), self.encoding))

    return line.strip()

  def ParseFileObject(self, parser_mediator, file_object, **kwargs):
    """Parses a text file-like object using a pyparsing definition.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      file_object (dfvfs.FileIO): file-like object.

    Raises:
      UnableToParseFile: when the file cannot be parsed.
    """
    # TODO: self._line_structures is a work-around and this needs
    # a structural fix.
    if not self._line_structures:
      raise errors.UnableToParseFile(
          u'Line structure undeclared, unable to proceed.')

    text_file_object = text_file.TextFile(file_object)

    line = self._ReadLine(
        parser_mediator, text_file_object, max_len=self.MAX_LINE_LENGTH,
        quiet=True)
    if not line:
      raise errors.UnableToParseFile(u'Not a text file.')

    if len(line) == self.MAX_LINE_LENGTH or len(
        line) == self.MAX_LINE_LENGTH - 1:
      logging.debug((
          u'Trying to read a line and reached the maximum allowed length of '
          u'{0:d}. The last few bytes of the line are: {1:s} [parser '
          u'{2:s}]').format(
              self.MAX_LINE_LENGTH, repr(line[-10:]), self.NAME))

    if not utils.IsText(line):
      raise errors.UnableToParseFile(u'Not a text file, unable to proceed.')

    if not self.VerifyStructure(parser_mediator, line):
      raise errors.UnableToParseFile(u'Wrong file structure.')

    # Set the offset to the beginning of the file.
    self._current_offset = 0
    # Read every line in the text file.
    while line:
      if parser_mediator.abort:
        break
      parsed_structure = None
      use_key = None
      # Try to parse the line using all the line structures.
      for key, structure in self.LINE_STRUCTURES:
        try:
          parsed_structure = structure.parseString(line)
        except pyparsing.ParseException:
          pass
        if parsed_structure:
          use_key = key
          break

      if parsed_structure:
        parsed_event = self.ParseRecord(
            parser_mediator, use_key, parsed_structure)
        if parsed_event:
          parsed_event.offset = self._current_offset
          parser_mediator.ProduceEvent(parsed_event)
      else:
        if len(line) > 80:
          line = u'{0:s}...'.format(line[:77])
        parser_mediator.ProduceExtractionError(
            u'unable to parse log line: {0:s} at offset {1:d}'.format(
                repr(line), self._current_offset))

      self._current_offset = text_file_object.get_offset()
      line = self._ReadLine(parser_mediator, text_file_object)

  @abc.abstractmethod
  def ParseRecord(self, parser_mediator, key, structure):
    """Parses a log record structure and produces events.

    This function takes as an input a parsed pyparsing structure
    and produces an EventObject if possible from that structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Returns:
      EventObject: event or None.
    """

  @abc.abstractmethod
  def VerifyStructure(self, parser_mediator, line):
    """Verify the structure of the file and return boolean based on that check.

    This function should read enough text from the text file to confirm
    that the file is the correct one for this particular parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      line (str): single line from the text file.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """


class EncodedTextReader(object):
  """Class to read simple encoded text."""

  def __init__(self, buffer_size=2048, encoding=None):
    """Initializes the encoded test reader object.

    Args:
      buffer_size (Optional[int]): buffer size.
      encoding (Optional[str]): encoding.
    """
    super(EncodedTextReader, self).__init__()
    self._buffer = b''
    self._buffer_size = buffer_size
    self._current_offset = 0
    self._encoding = encoding

    if self._encoding:
      self._new_line = u'\n'.encode(self._encoding)
      self._carriage_return = u'\r'.encode(self._encoding)
    else:
      self._new_line = b'\n'
      self._carriage_return = b'\r'

    self._new_line_length = len(self._new_line)
    self._carriage_return_length = len(self._carriage_return)

    self.lines = u''

  def _ReadLine(self, file_object):
    """Reads a line from the file object.

    Args:
      file_object (dfvfs.FileIO): file-like object.

    Returns:
      str: line read from the file-like object.
    """
    if len(self._buffer) < self._buffer_size:
      self._buffer = b''.join([
          self._buffer, file_object.read(self._buffer_size)])

    line, new_line, self._buffer = self._buffer.partition(self._new_line)
    if not line and not new_line:
      line = self._buffer
      self._buffer = b''

    self._current_offset += len(line)

    # Strip carriage returns from the text.
    if line.endswith(self._carriage_return):
      line = line[:-self._carriage_return_length]

    if new_line:
      line = b''.join([line, self._new_line])
      self._current_offset += self._new_line_length

    # If a parser specifically indicates specific encoding we need
    # to handle the buffer as it is an encoded string.
    # If it fails we fail back to the original raw string.
    if self._encoding:
      try:
        line = line.decode(self._encoding)
      except UnicodeDecodeError:
        # TODO: it might be better to raise here.
        pass

    return line

  def ReadLine(self, file_object):
    """Reads a line.

    Args:
      file_object (dfvfs.FileIO): file-like object.

    Returns:
      str: line read from the lines buffer.
    """
    line, _, self.lines = self.lines.partition(u'\n')
    if not line:
      self.ReadLines(file_object)
      line, _, self.lines = self.lines.partition(u'\n')

    return line

  def ReadLines(self, file_object):
    """Reads lines into the lines buffer.

    Args:
      file_object (dfvfs.FileIO): file-like object.
    """
    lines_size = len(self.lines)
    if lines_size < self._buffer_size:
      lines_size = self._buffer_size - lines_size
      while lines_size > 0:
        line = self._ReadLine(file_object)
        if not line:
          break

        self.lines = u''.join([self.lines, line])
        lines_size -= len(line)

  def Reset(self):
    """Resets the encoded text reader."""
    self._buffer = b''
    self._current_offset = 0

    self.lines = u''

  def SkipAhead(self, file_object, number_of_characters):
    """Skips ahead a number of characters.

    Args:
      file_object (dfvfs.FileIO): file-like object.
      number_of_characters (int): number of characters.
    """
    lines_size = len(self.lines)
    while number_of_characters >= lines_size:
      number_of_characters -= lines_size

      self.lines = u''
      self.ReadLines(file_object)
      lines_size = len(self.lines)
      if lines_size == 0:
        return

    self.lines = self.lines[number_of_characters:]


class PyparsingMultiLineTextParser(PyparsingSingleLineTextParser):
  """Multi line text parser based on the pyparsing library."""

  BUFFER_SIZE = 2048

  def __init__(self):
    """Initializes a parser object."""
    super(PyparsingMultiLineTextParser, self).__init__()
    self._buffer_size = self.BUFFER_SIZE
    self._text_reader = EncodedTextReader(
        buffer_size=self.BUFFER_SIZE, encoding=self._ENCODING)

  def ParseFileObject(self, parser_mediator, file_object, **kwargs):
    """Parses a text file-like object using a pyparsing definition.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      file_object (dfvfs.FileIO): file-like object.

    Raises:
      UnableToParseFile: when the file cannot be parsed.
    """
    if not self.LINE_STRUCTURES:
      raise errors.UnableToParseFile(u'Missing line structures.')

    self._text_reader.Reset()

    try:
      self._text_reader.ReadLines(file_object)
    except UnicodeDecodeError as exception:
      raise errors.UnableToParseFile(
          u'Not a text file, with error: {0:s}'.format(exception))

    if not utils.IsText(self._text_reader.lines):
      raise errors.UnableToParseFile(u'Not a text file, unable to proceed.')

    if not self.VerifyStructure(parser_mediator, self._text_reader.lines):
      raise errors.UnableToParseFile(u'Wrong file structure.')

    # Using parseWithTabs() overrides Pyparsing's default replacement of tabs
    # with spaces to SkipAhead() the correct number of bytes after a match.
    for key, structure in self.LINE_STRUCTURES:
      structure.parseWithTabs()

    # Read every line in the text file.
    while self._text_reader.lines:
      if parser_mediator.abort:
        break

      # Initialize pyparsing objects.
      tokens = None
      start = 0
      end = 0

      key = None

      # Try to parse the line using all the line structures.
      for key, structure in self.LINE_STRUCTURES:
        try:
          structure_generator = structure.scanString(
              self._text_reader.lines, maxMatches=1)
          parsed_structure = next(structure_generator, None)
        except pyparsing.ParseException:
          parsed_structure = None

        if not parsed_structure:
          continue

        tokens, start, end = parsed_structure

        # Only want to parse the structure if it starts
        # at the beginning of the buffer.
        if start == 0:
          break

      if tokens and start == 0:
        try:
          self.ParseRecord(parser_mediator, key, tokens)
        except (errors.ParseError, errors.TimestampError) as exception:
          parser_mediator.ProduceExtractionError(
              u'unable parse record: {0:s} with error: {1:s}'.format(
                  key, exception))

        self._text_reader.SkipAhead(file_object, end)

      else:
        odd_line = self._text_reader.ReadLine(file_object)
        if odd_line:
          if len(odd_line) > 80:
            odd_line = u'{0:s}...'.format(odd_line[:77])
          parser_mediator.ProduceExtractionError(
              u'unable to parse log line: {0:s}'.format(repr(odd_line)))

      try:
        self._text_reader.ReadLines(file_object)
      except UnicodeDecodeError as exception:
        parser_mediator.ProduceExtractionError(
            u'unable to read lines with error: {0:s}'.format(exception))

  @abc.abstractmethod
  def ParseRecord(self, parser_mediator, key, structure):
    """Parses a log record structure and produces events.

    This function takes as an input a parsed pyparsing structure
    and produces an EventObject if possible from that structure.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      key (str): name of the parsed structure.
      structure (pyparsing.ParseResults): tokens from a parsed log line.

    Returns:
      EventObject: event or None.
    """

  @abc.abstractmethod
  def VerifyStructure(self, parser_mediator, line):
    """Verify the structure of the file and return boolean based on that check.

    This function should read enough text from the text file to confirm
    that the file is the correct one for this particular parser.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      line (str): single line from the text file.

    Returns:
      bool: True if this is the correct parser, False otherwise.
    """



# -*- coding: utf-8 -*-
"""This file contains the definitions for analysis plugins."""

# All the possible analysis plugin types.

# Plugin that detects anomalies.
PLUGIN_TYPE_ANOMALY = 1

# Plugin that calculating statistical information.
PLUGIN_TYPE_STATISTICS = 2

# Plugin that annotates (or tags) event objects.
PLUGIN_TYPE_ANNOTATION = 3

# Plugin that provides summary information.
PLUGIN_TYPE_REPORT = 4



# -*- coding: utf-8 -*-
"""The task manager."""

import collections
import heapq
import logging
import threading
import time

from dfvfs.lib import definitions as dfvfs_definitions

from plaso.containers import tasks
from plaso.engine import processing_status


class _PendingMergeTaskHeap(object):
  """Heap to manage pending merge tasks."""

  def __init__(self):
    """Initializes a pending merge task heap."""
    super(_PendingMergeTaskHeap, self).__init__()
    self._heap = []
    self._task_identifiers = set()

  def __len__(self):
    """int: number of tasks on the heap."""
    return len(self._heap)

  def __contains__(self, task_identifier):
    """Checks for an task identifier being present in the heap.

    Args:
      task_identifier (str): task identifier to check for.

    Returns:
      bool: True if the task with the given identifier is present in the heap.
    """
    return task_identifier in self._task_identifiers

  def PeekTask(self):
    """Retrieves the first task from the heap without removing it.

    Returns:
      Task: task or None if the heap is empty.
    """
    try:
      _, task = self._heap[0]

    except IndexError:
      return

    return task

  def PopTask(self):
    """Retrieves and removes the first task from the heap.

    Returns:
      Task: the task or None if the heap is empty.
    """
    try:
      _, task = heapq.heappop(self._heap)

    except IndexError:
      return
    self._task_identifiers.remove(task.identifier)
    return task

  def PushTask(self, task):
    """Pushes a task onto the heap.

    Args:
      task (Task): task.

    Raises:
      ValueError: if the size of the storage file is not set in the task.
    """
    storage_file_size = getattr(task, u'storage_file_size', None)
    if not storage_file_size:
      raise ValueError(u'Task storage file size not set')

    if task.file_entry_type == dfvfs_definitions.FILE_ENTRY_TYPE_DIRECTORY:
      weight = 1
    else:
      weight = storage_file_size

    task.merge_priority = weight

    heap_values = (weight, task)
    heapq.heappush(self._heap, heap_values)
    self._task_identifiers.add(task.identifier)


class TaskManager(object):
  """Manages tasks and tracks their completion and status.

  A task being tracked by the manager must be in exactly one of the
  following states:

  * abandoned: no status information has been recently received from a worker
      about the task, and is assumed to be abandoned.
  * queued: the task is waiting for a worker to start processing it. It's also
      possible that a worker has already completed the task, but no status
      update was collected from the worker while it processed the task.
  * processing: a worker is processing the task.
  * pending_merge: a worker has completed processing the task and the
      results are ready to be merged with the session storage.
  * merging: tasks that are being merged by the engine.

  Once the engine reports that a task is completely merged, it is removed
  from the task manager.

  Tasks that are not abandoned, or abandoned, but need to be retried are
  considered "pending", as there is more work that needs to be done to complete
  them.
  """

  # Consider a task inactive after 5 minutes of no activity.
  _TASK_INACTIVE_TIME = 5 * 60 * 1000000

  def __init__(self):
    """Initializes a task manager."""
    super(TaskManager, self).__init__()
    self._lock = threading.Lock()
    # This dictionary maps task identifiers to tasks that have been abandoned,
    # as no worker has reported processing the task in the expected interval.
    self._tasks_abandoned = {}
    self._tasks_pending_merge = _PendingMergeTaskHeap()
    # This dictionary maps task identifiers to tasks that are currently
    # being processed by a worker.
    self._tasks_processing = {}
    # Use ordered dictionary to preserve the order in which tasks were added.
    # This dictionary maps task identifiers to tasks that are waiting to be
    # processed.
    self._tasks_queued = collections.OrderedDict()
    # This dictionary maps task identifiers to tasks that are being merged
    # by the foreman.
    self._tasks_merging = {}

    # TODO: implement a limit on the number of tasks.
    self._total_number_of_tasks = 0

  # TODO: add support for task types.
  def CreateTask(self, session_identifier):
    """Creates a task.

    Args:
      session_identifier (str): the identifier of the session the task is
          part of.

    Returns:
      Task: task attribute container.
    """
    task = tasks.Task(session_identifier)
    with self._lock:
      self._tasks_queued[task.identifier] = task

    self._total_number_of_tasks += 1
    return task

  def CompleteTask(self, task):
    """Completes a task.

    The task is complete and can be removed from the task manager.

    Args:
      task (Task): task.
    """
    with self._lock:
      if task.identifier in self._tasks_merging:
        del self._tasks_merging[task.identifier]
        logging.debug(u'Task {0:s} is complete.'.format(task.identifier))

      if task.identifier in self._tasks_pending_merge:
        logging.debug(u'Task {0:s} completed while pending merge.'.format(
            task.identifier))
        return

      if task.identifier in self._tasks_processing:
        del self._tasks_processing[task.identifier]
        logging.debug(u'Task {0:s} completed from processing'.format(
            task.identifier))
        return

      if task.identifier in self._tasks_queued:
        del self._tasks_queued[task.identifier]
        logging.debug(u'Task {0:s} is completed from queued'.format(
            task.identifier))
        return

  def GetAbandonedTasks(self):
    """Retrieves all abandoned tasks.

    Returns:
      list[Task]: tasks.
    """
    with self._lock:
      abandoned_tasks = list(self._tasks_abandoned.values())
    return abandoned_tasks

  def GetTaskPendingMerge(self, current_task):
    """Retrieves the first task that is pending merge or has a higher priority.

    This function will check if there is a task with a higher merge priority
    than the current_task being merged. If so, that task with the higher
    priority is returned.

    Args:
      current_task (Task): current task being merged or None if no such task.

    Returns:
      Task: the next task to merge or None if there is no task pending merge or
          with a higher priority.
    """
    next_task = self._tasks_pending_merge.PeekTask()
    if not next_task:
      return

    if current_task and next_task.merge_priority > current_task.merge_priority:
      return

    with self._lock:
      next_task = self._tasks_pending_merge.PopTask()
    self._tasks_merging[next_task.identifier] = next_task
    return next_task

  def GetStatusInformation(self):
    """Retrieves status information about the tasks.

    Returns:
      TasksStatus: tasks status information.
    """
    status = processing_status.TasksStatus()

    with self._lock:
      status.number_of_abandoned_tasks = len(self._tasks_abandoned)
      status.number_of_queued_tasks = len(self._tasks_queued)
      status.number_of_tasks_pending_merge = len(self._tasks_pending_merge)
      status.number_of_tasks_processing = len(self._tasks_processing)
      status.total_number_of_tasks = self._total_number_of_tasks

    return status

  def GetTasksCheckMerge(self):
    """Retrieves the tasks that need to be checked if they are ready for merge.

    Returns:
      list[Task]: tasks that are being processed by workers or that have been
          abandoned.
    """
    with self._lock:
      tasks_list = list(self._tasks_processing.values())
      tasks_list.extend(self._tasks_queued.values())
      tasks_list.extend(self._tasks_abandoned.values())
    return tasks_list

  def _TimeoutTasks(self, tasks_for_timeout):
    """Checks for inactive tasks, and marks such tasks as abandoned.

    Note that this method does not lock the manager, and should be called
    by a method holding the manager lock.

    Args:
      tasks_for_timeout (dict[str, Task]): mapping of task identifiers to Tasks
        that will be checked for inactivity, and marked as abandoned if
        required.
    """
    if not tasks_for_timeout:
      return
    inactive_time = int(time.time() * 1000000) - self._TASK_INACTIVE_TIME

    for task_identifier, task in iter(tasks_for_timeout.items()):
      last_active_time = task.last_processing_time
      if not last_active_time:
        last_active_time = task.start_time
      if last_active_time < inactive_time:
        logging.debug(u'Task {0:s} is abandoned'.format(task_identifier))
        self._tasks_abandoned[task_identifier] = task
        del tasks_for_timeout[task_identifier]

  def HasPendingTasks(self):
    """Determines if there are tasks running, or in need of retrying.

    Returns:
      bool: True if there are tasks that are active, ready to be merged, or
          need to be retried.
    """
    with self._lock:
      self._TimeoutTasks(self._tasks_processing)
      if self._tasks_processing:
        return True
      # There are no tasks being processed, but we might be
      # waiting for some tasks to be merged.
      if self._HasTasksPendingMerge():
        return True
      # There are no tasks processing, or pending merge, but there may
      # still be some waiting to be retried, so we check that.
      if self._HasTasksPendingRetry():
        return True
      # It's possible that a worker received a task and didn't report it.
      # Check for tasks that we believe are queued that might need to be marked
      # as abandoned, as all the workers are idle, and the task has been queued
      # for longer than the timeout.
      self._TimeoutTasks(self._tasks_queued)
      if self._tasks_queued:
        return True
      if self._tasks_merging:
        return True

    # There are no tasks pending any work.
    return False

  def _TaskIsRetriable(self, task):
    """Determines if a task is eligible to be retried.

    Args:
      task (Task): task to be checked for its eligibility to be retried.

    Returns:
      bool: True if the task is eligible to be retried.
    """
    return not (task.retried or task.original_task_identifier)

  def _HasTasksPendingMerge(self):
    """Determines if there are tasks waiting to be merged.

    Note that this method does not lock the manager, and should be called
    by a method holding the manager lock.

    Returns:
      bool: True if there are abandoned tasks that need to be retried.
    """
    return bool(self._tasks_pending_merge)

  def _HasTasksPendingRetry(self):
    """Determines if there are abandoned tasks that still need to be retried.

    Note that this method does not lock the manager, and should be called
    by a method holding the manager lock.

    Returns:
      bool: True if there are abandoned tasks that need to be retried.
    """
    for abandoned_task in self._tasks_abandoned.values():
      if self._TaskIsRetriable(abandoned_task):
        return True

  def GetRetryTask(self):
    """Creates a task that is an attempt to retry an abandoned task.

    Returns:
      Task: a task that is a retry of an existing task, or None if there are
          no tasks that need to be retried.
    """
    with self._lock:
      for abandoned_task in self._tasks_abandoned.values():
        # Only retry abandoned tasks that are yet to be retried, and
        # aren't themselves retries of another task.
        if self._TaskIsRetriable(abandoned_task):
          retry_task = abandoned_task.CreateRetry()
          logging.debug(
              u'Retrying task {0:s} as {1:s}'.format(
                  abandoned_task.identifier, retry_task.identifier))
          self._tasks_queued[retry_task.identifier] = retry_task
          self._total_number_of_tasks += 1
          return retry_task

  def UpdateTaskAsPendingMerge(self, task):
    """Updates the task manager to reflect the task is ready to be merged.

    Args:
      task (Task): task.

    Raises:
      KeyError: if the task was not processing or abandoned.
    """
    with self._lock:
      is_processing = task.identifier in self._tasks_processing
      is_abandoned = task.identifier in self._tasks_abandoned
      is_queued = task.identifier in self._tasks_queued

      if not (is_queued or is_abandoned or is_processing):
        raise KeyError(u'Status of task {0:s} is unknown'.format(
            task.identifier))

      self._tasks_pending_merge.PushTask(task)
      task.UpdateProcessingTime()

      if is_queued:
        del self._tasks_queued[task.identifier]

      if is_processing:
        del self._tasks_processing[task.identifier]

      if is_abandoned:
        del self._tasks_abandoned[task.identifier]

    if is_abandoned:
      logging.warning(
          u'Previously abandoned task {0:s} is now pending merge'.format(
              task.identifier))
    else:
      logging.debug(u'Task {0:s} is pending merge'.format(task.identifier))

  def UpdateTaskAsProcessingByIdentifier(self, task_identifier):
    """Updates the task manager to reflect the task is processing.

    Args:
      task_identifier (str): unique identifier of the task.

    Raises:
      KeyError: if the task is not known to the task manager.
    """
    with self._lock:
      task_processing = self._tasks_processing.get(task_identifier, None)
      if task_processing:
        task_processing.UpdateProcessingTime()
        return

      task_queued = self._tasks_queued.get(task_identifier, None)
      if task_queued:
        logging.debug(u'Task {0:s} was queued, now processing'.format(
            task_identifier))
        self._tasks_processing[task_identifier] = task_queued
        del self._tasks_queued[task_identifier]
        task_queued.UpdateProcessingTime()
        return

      task_abandoned = self._tasks_abandoned.get(task_identifier, None)
      if task_abandoned:
        del self._tasks_abandoned[task_identifier]
        self._tasks_processing[task_identifier] = task_abandoned
        logging.debug(u'Task {0:s} was abandoned, but now processing'.format(
            task_identifier))
        task_abandoned.UpdateProcessingTime()
        return

      if task_identifier in self._tasks_pending_merge:
        # No need to update the processing time, as this task is already
        # has already finished processing, and is just waiting for merge.
        return

    # If we get here, we don't know what state the tasks is in, so raise.
    raise KeyError(u'Status of task {0:s} is unknown'.format(task_identifier))



# -*- coding: utf-8 -*-
# There's a backslash in the module docstring, so as not to confuse Sphinx.
# pylint: disable=anomalous-backslash-in-string
"""This file contains XChat log file parser in plaso.

Information updated 24 July 2013.

The parser applies to XChat log files. Despite their apparent
simplicity it's not straightforward to manage every possible case.
XChat tool allows users to specify how timestamp will be
encoded (using the strftime function), by letting them to specify
additional separators. This parser will accept only the simplest
default English form of an XChat log file, as the following::

  **** BEGIN LOGGING AT Mon Dec 31 21:11:55 2001
  dec 31 21:11:55 --> You are now talking on #gugle
  dec 31 21:11:55 --- Topic for #gugle is plaso, nobody knows what it means
  dec 31 21:11:55 Topic for #gugle set by Kristinn
  dec 31 21:11:55 --- Joachim gives voice to fpi
  dec 31 21:11:55 *   XChat here
  dec 31 21:11:58 <fpi> ola plas-ing guys!
  dec 31 21:12:00 <Kristinn> ftw!

It could be managed the missing month/day case too, by extracting
the month/day information from the header. But the parser logic
would become intricate, since it would need to manage day transition,
chat lines crossing the midnight. From there derives the last day of
the year bug, since the parser will not manage that transition.

Moreover the strftime is locale-dependant, so month names, footer and
headers can change, even inside the same log file. Being said that, the
following will be the main logic used to parse the log files (note that
the first header *must be* '\*\*\*\* BEGIN ...' otherwise file will be skipped).

1) Check for '\*\*\*\*'
1.1) If 'BEGIN LOGGING AT' (English)
1.1.1) Extract the YEAR
1.1.2) Generate new event start logging
1.1.3) set parsing = True
1.2) If 'END LOGGING'
1.2.1) If parsing, set parsing=False
1.2.2) If not parsing, log debug
1.2.3) Generate new event end logging
1.3) If not BEGIN|END we are facing a different language
and we don't now which language!
If parsing is True, set parsing=False and log debug
2) Not '\*\*\*\*' so we are parsing a line
2.1) If parsing = True, try to parse line and generate event
2.2) If parsing = False, skip until next good header is found

References
http://xchat.org
"""

import logging

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.containers import time_events
from plaso.lib import errors
from plaso.lib import definitions
from plaso.lib import timelib
from plaso.parsers import manager
from plaso.parsers import text_parser


__author__ = 'Francesco Picasso (francesco.picasso@gmail.com)'


class XChatLogEventData(events.EventData):
  """XChat Log event data.

  Attributes:
    nickname (str): nickname.
    text (str): text sent by nickname or other text (server, messages, etc.).
  """

  DATA_TYPE = u'xchat:log:line'

  def __init__(self):
    """Initializes event data."""
    super(XChatLogEventData, self).__init__(data_type=self.DATA_TYPE)
    self.nickname = None
    self.text = None


class XChatLogParser(text_parser.PyparsingSingleLineTextParser):
  """Parse XChat log files."""

  NAME = u'xchatlog'
  DESCRIPTION = u'Parser for XChat log files.'

  _ENCODING = u'UTF-8'

  # Common (header/footer/body) pyparsing structures.
  # TODO: Only English ASCII timestamp supported ATM, add support for others.

  _WEEKDAY = pyparsing.Group(
      pyparsing.Keyword(u'Sun') |
      pyparsing.Keyword(u'Mon') |
      pyparsing.Keyword(u'Tue') |
      pyparsing.Keyword(u'Wed') |
      pyparsing.Keyword(u'Thu') |
      pyparsing.Keyword(u'Fri') |
      pyparsing.Keyword(u'Sat'))

  # Header/footer pyparsing structures.
  # Sample: "**** BEGIN LOGGING AT Mon Dec 31 21:11:55 2011".
  # Note that "BEGIN LOGGING" text is localized (default, English) and can be
  # different if XChat locale is different.

  _HEADER_SIGNATURE = pyparsing.Keyword(u'****')
  _HEADER_DATE_TIME = pyparsing.Group(
      _WEEKDAY.setResultsName(u'weekday') +
      text_parser.PyparsingConstants.THREE_LETTERS.setResultsName(u'month') +
      text_parser.PyparsingConstants.ONE_OR_TWO_DIGITS.setResultsName(u'day') +
      text_parser.PyparsingConstants.TIME_ELEMENTS +
      text_parser.PyparsingConstants.FOUR_DIGITS.setResultsName(u'year'))
  _LOG_ACTION = pyparsing.Group(
      pyparsing.Word(pyparsing.printables) +
      pyparsing.Word(pyparsing.printables) +
      pyparsing.Word(pyparsing.printables))
  _HEADER = (
      _HEADER_SIGNATURE.suppress() + _LOG_ACTION.setResultsName(u'log_action') +
      _HEADER_DATE_TIME.setResultsName(u'date_time'))

  # Body (nickname, text and/or service messages) pyparsing structures.
  # Sample: "dec 31 21:11:58 <fpi> ola plas-ing guys!".

  _DATE_TIME = pyparsing.Group(
      text_parser.PyparsingConstants.THREE_LETTERS.setResultsName(u'month') +
      text_parser.PyparsingConstants.ONE_OR_TWO_DIGITS.setResultsName(u'day') +
      text_parser.PyparsingConstants.TIME_ELEMENTS)
  _NICKNAME = pyparsing.QuotedString(u'<', endQuoteChar=u'>').setResultsName(
      u'nickname')
  _LOG_LINE = (
      _DATE_TIME.setResultsName(u'date_time') +
      pyparsing.Optional(_NICKNAME) +
      pyparsing.SkipTo(pyparsing.lineEnd).setResultsName(u'text'))

  LINE_STRUCTURES = [
      (u'logline', _LOG_LINE),
      (u'header', _HEADER),
      (u'header_signature', _HEADER_SIGNATURE),
  ]

  def __init__(self):
    """Initializes a parser object."""
    super(XChatLogParser, self).__init__()
    self._last_month = 0
    self._xchat_year = None
    self.offset = 0

  def _GetTimeElementsTuple(self, structure):
    """Retrieves a time elements tuple from the structure.

    Args:
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.

    Returns:
      tuple: contains:
        year (int): year.
        month (int): month, where 1 represents January.
        day_of_month (int): day of month, where 1 is the first day of the month.
        hours (int): hours.
        minutes (int): minutes.
        seconds (int): seconds.
    """
    month, day, hours, minutes, seconds = structure.date_time

    month = timelib.MONTH_DICT.get(month.lower(), 0)

    if month != 0 and month < self._last_month:
      # Gap detected between years.
      self._xchat_year += 1

    return (self._xchat_year, month, day, hours, minutes, seconds)

  def _ParseHeader(self, parser_mediator, structure):
    """Parses a log header.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.
    """
    _, month, day, hours, minutes, seconds, year = structure.date_time

    month = timelib.MONTH_DICT.get(month.lower(), 0)

    time_elements_tuple = (year, month, day, hours, minutes, seconds)

    try:
      date_time = dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)
      date_time.is_local_time = True
    except ValueError:
      parser_mediator.ProduceExtractionError(
          u'invalid date time value: {0!s}'.format(structure.date_time))
      return

    self._last_month = month

    event_data = XChatLogEventData()

    if structure.log_action[0] == u'BEGIN':
      self._xchat_year = year
      event_data.text = u'XChat start logging'

    elif structure.log_action[0] == u'END':
      self._xchat_year = None
      event_data.text = u'XChat end logging'

    else:
      logging.debug(u'Unknown log action: {0:s}.'.format(
          u' '.join(structure.log_action)))
      return

    event = time_events.DateTimeValuesEvent(
        date_time, definitions.TIME_DESCRIPTION_ADDED,
        time_zone=parser_mediator.timezone)
    parser_mediator.ProduceEventWithEventData(event, event_data)

  def _ParseLogLine(self, parser_mediator, structure):
    """Parses a log line.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.
    """
    if not self._xchat_year:
      return

    time_elements_tuple = self._GetTimeElementsTuple(structure)

    try:
      date_time = dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)
      date_time.is_local_time = True
    except ValueError:
      parser_mediator.ProduceExtractionError(
          u'invalid date time value: {0!s}'.format(structure.date_time))
      return

    self._last_month = time_elements_tuple[1]

    event_data = XChatLogEventData()
    event_data.nickname = structure.nickname
    # The text string contains multiple unnecessary whitespaces that need to
    # be removed, thus the split and re-join.
    event_data.text = u' '.join(structure.text.split())

    event = time_events.DateTimeValuesEvent(
        date_time, definitions.TIME_DESCRIPTION_ADDED,
        time_zone=parser_mediator.timezone)
    parser_mediator.ProduceEventWithEventData(event, event_data)

  def ParseRecord(self, parser_mediator, key, structure):
    """Parses a log record structure and produces events.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      key (str): identifier of the structure of tokens.
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.

    Raises:
      ParseError: when the structure type is unknown.
    """
    if key not in (u'header', u'header_signature', u'logline'):
      raise errors.ParseError(
          u'Unable to parse record, unknown structure: {0:s}'.format(key))

    if key == u'logline':
      self._ParseLogLine(parser_mediator, structure)

    elif key == u'header':
      self._ParseHeader(parser_mediator, structure)

    elif key == u'header_signature':
      # If this key is matched (after others keys failed) we got a different
      # localized header and we should stop parsing until a new good header
      # is found. Stop parsing is done setting xchat_year to 0.
      # Note that the code assumes that LINE_STRUCTURES will be used in the
      # exact order as defined!
      logging.warning(u'Unknown locale header.')
      self._xchat_year = 0

  def VerifyStructure(self, parser_mediator, line):
    """Verify that this file is a XChat log file.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      line (bytes): line from a text file.

    Returns:
      bool: True if the line is in the expected format, False if not.
    """
    try:
      structure = self._HEADER.parseString(line)
    except pyparsing.ParseException:
      logging.debug(u'Not a XChat log file')
      return False

    _, month, day, hours, minutes, seconds, year = structure.date_time

    month = timelib.MONTH_DICT.get(month.lower(), 0)

    time_elements_tuple = (year, month, day, hours, minutes, seconds)

    try:
      dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)
    except ValueError:
      logging.debug(
          u'Not a XChat log file, invalid date and time: {0!s}'.format(
              structure.date_time))
      return False

    return True


manager.ParsersManager.RegisterParser(XChatLogParser)



# -*- coding: utf-8 -*-
"""This file contains a plugin for SSH syslog entries."""

import pyparsing

from plaso.containers import time_events
from plaso.lib import definitions
from plaso.parsers import syslog
from plaso.parsers.syslog_plugins import interface


class CronTaskRunEventData(syslog.SyslogLineEventData):
  """Cron task run event data.

  Attributes:
    command (str): command executed.
    username (str): name of user the command was executed.
  """

  DATA_TYPE = u'syslog:cron:task_run'

  def __init__(self):
    """Initializes event data."""
    super(CronTaskRunEventData, self).__init__(data_type=self.DATA_TYPE)
    self.command = None
    self.username = None


class CronPlugin(interface.SyslogPlugin):
  """A syslog plugin for parsing cron messages."""
  NAME = u'cron'

  DESCRIPTION = u'Parser for syslog cron messages.'

  REPORTER = u'CRON'

  _PYPARSING_COMPONENTS = {
      u'command': pyparsing.Combine(
          pyparsing.SkipTo(
              pyparsing.Literal(u')') + pyparsing.StringEnd())).
                  setResultsName(u'command'),
      u'username': pyparsing.Word(pyparsing.alphanums).setResultsName(
          u'username'),
  }

  _TASK_RUN_GRAMMAR = (
      pyparsing.Literal(u'(') + _PYPARSING_COMPONENTS[u'username'] +
      pyparsing.Literal(u')') + pyparsing.Literal(u'CMD') +
      pyparsing.Literal(u'(') + _PYPARSING_COMPONENTS[u'command'] +
      pyparsing.Literal(u')') + pyparsing.StringEnd()
  )

  MESSAGE_GRAMMARS = [(u'task_run', _TASK_RUN_GRAMMAR)]

  def ParseMessage(self, parser_mediator, key, timestamp, tokens):
    """Parses a syslog body that matched one of defined grammars.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      key (str): name of the matching grammar.
      timestamp (int): the timestamp, which contains the number of micro seconds
          since January 1, 1970, 00:00:00 UTC or 0 on error.
      tokens (dict[str, str]): tokens derived from a syslog message based on
          the defined grammar.

    Raises:
      AttributeError: If an unknown key is provided.
    """
    # TODO: change AttributeError into ValueError or equiv.
    if key != u'task_run':
      raise AttributeError(u'Unknown grammar key: {0:s}'.format(key))

    event_data = CronTaskRunEventData()
    event_data.body = tokens.get(u'body', None)
    event_data.command = tokens.get(u'command', None)
    event_data.hostname = tokens.get(u'hostname', None)
    # TODO: pass line number to offset or remove.
    event_data.offset = 0
    event_data.pid = tokens.get(u'pid', None)
    event_data.reporter = tokens.get(u'reporter', None)
    event_data.severity = tokens.get(u'severity', None)
    event_data.username = tokens.get(u'username', None)

    event = time_events.TimestampEvent(
        timestamp, definitions.TIME_DESCRIPTION_WRITTEN)
    parser_mediator.ProduceEventWithEventData(event, event_data)


syslog.SyslogParser.RegisterPlugin(CronPlugin)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Tests for the Networks Windows Registry plugin."""

import unittest

from dfdatetime import filetime as dfdatetime_filetime
from dfwinreg import definitions as dfwinreg_definitions
from dfwinreg import fake as dfwinreg_fake

from plaso.formatters import winreg  # pylint: disable=unused-import
from plaso.lib import definitions
from plaso.lib import timelib
from plaso.parsers.winreg_plugins import networks

from tests.parsers.winreg_plugins import test_lib


class NetworksPluginTest(test_lib.RegistryPluginTestCase):
  """Tests for the Networks Windows Registry plugin."""

  def _CreateTestKey(self, key_path, time_string):
    """Creates Registry keys and values for testing.

    Args:
      key_path (str): Windows Registry key path.
      time_string (str): key last written date and time.

    Returns:
      dfwinreg.WinRegistryKey: a Windows Registry key.
    """
    filetime = dfdatetime_filetime.Filetime()
    filetime.CopyFromString(time_string)
    registry_key = dfwinreg_fake.FakeWinRegistryKey(
        u'NetworkList', key_path=key_path,
        last_written_time=filetime.timestamp, offset=153)

    # Setup Profiles.
    profiles = dfwinreg_fake.FakeWinRegistryKey(u'Profiles')
    registry_key.AddSubkey(profiles)

    profile_1 = dfwinreg_fake.FakeWinRegistryKey(
        u'{B358E985-4464-4ABD-AF99-7D4A0AF66BB7}')
    profiles.AddSubkey(profile_1)

    value_data = b'\x00\x00\x00\x00'
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'Category', data=value_data,
        data_type=dfwinreg_definitions.REG_DWORD_BIG_ENDIAN)
    profile_1.AddValue(registry_value)

    value_data = (
        b'\xde\x07\x0c\x00\x02\x00\x10\x00\x08\x00\x04\x00\x27\x00\x6a\x00')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'DateCreated', data=value_data,
        data_type=dfwinreg_definitions.REG_BINARY)
    profile_1.AddValue(registry_value)

    value_data = (
        b'\xdf\x07\x01\x00\x02\x00\x1b\x00\x0f\x00\x0f\x00\x1b\x00\xc5\x03')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'DateLastConnected', data=value_data,
        data_type=dfwinreg_definitions.REG_BINARY)
    profile_1.AddValue(registry_value)

    value_data = u'My Awesome Wifi Hotspot'.encode(u'utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'Description', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ)
    profile_1.AddValue(registry_value)

    value_data = b'\x00\x00\x00\x00'
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'Managed', data=value_data,
        data_type=dfwinreg_definitions.REG_DWORD_BIG_ENDIAN)
    profile_1.AddValue(registry_value)

    value_data = b'\x00\x00\x00\x47'
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'NameType', data=value_data,
        data_type=dfwinreg_definitions.REG_DWORD_BIG_ENDIAN)
    profile_1.AddValue(registry_value)

    value_data = u'My Awesome Wifi Hotspot'.encode(u'utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'ProfileName', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ)
    profile_1.AddValue(registry_value)

    profile_2 = dfwinreg_fake.FakeWinRegistryKey(
        u'{C1C57B58-BFE2-428B-818C-9D69A873AD3D}')
    profiles.AddSubkey(profile_2)

    value_data = b'\x00\x00\x00\x00'
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'Category', data=value_data,
        data_type=dfwinreg_definitions.REG_DWORD_BIG_ENDIAN)
    profile_2.AddValue(registry_value)

    value_data = (
        b'\xde\x07\x05\x00\x02\x00\x06\x00\x11\x00\x02\x00\x13\x00\x1b\x03')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'DateCreated', data=value_data,
        data_type=dfwinreg_definitions.REG_BINARY)
    profile_2.AddValue(registry_value)

    value_data = (
        b'\xde\x07\x05\x00\x02\x00\x06\x00\x11\x00\x07\x00\x36\x00\x0a\x00')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'DateLastConnected', data=value_data,
        data_type=dfwinreg_definitions.REG_BINARY)
    profile_2.AddValue(registry_value)

    value_data = u'Network'.encode(u'utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'Description', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ)
    profile_2.AddValue(registry_value)

    value_data = b'\x00\x00\x00\x00'
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'Managed', data=value_data,
        data_type=dfwinreg_definitions.REG_DWORD_BIG_ENDIAN)
    profile_2.AddValue(registry_value)

    value_data = b'\x00\x00\x00\x06'
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'NameType', data=value_data,
        data_type=dfwinreg_definitions.REG_DWORD_BIG_ENDIAN)
    profile_2.AddValue(registry_value)

    value_data = u'Network'.encode(u'utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'ProfileName', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ)
    profile_2.AddValue(registry_value)

    # Setup signatures.
    signatures = dfwinreg_fake.FakeWinRegistryKey(u'Signatures')
    registry_key.AddSubkey(signatures)

    managed = dfwinreg_fake.FakeWinRegistryKey(u'Managed')
    signatures.AddSubkey(managed)

    unmanaged = dfwinreg_fake.FakeWinRegistryKey(u'Unmanaged')
    signatures.AddSubkey(unmanaged)

    unmanaged_subkey = dfwinreg_fake.FakeWinRegistryKey(
        u'010103000F0000F0080000000F0000F0E8982FB31F37E52AF30A6575A4898CE667'
        u'6E8C2F99C4C5131D84F64BD823E0')
    unmanaged.AddSubkey(unmanaged_subkey)

    value_data = b'\x00\x50\x56\xea\x6c\xec'
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'DefaultGatewayMac', data=value_data,
        data_type=dfwinreg_definitions.REG_BINARY)
    unmanaged_subkey.AddValue(registry_value)

    value_data = u'Network'.encode(u'utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'Description', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ)
    unmanaged_subkey.AddValue(registry_value)

    value_data = u'localdomain'.encode(u'utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'DnsSuffix', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ)
    unmanaged_subkey.AddValue(registry_value)

    value_data = u'Network'.encode(u'utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'FirstNetwork', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ)
    unmanaged_subkey.AddValue(registry_value)

    value_data = u'{C1C57B58-BFE2-428B-818C-9D69A873AD3D}'.encode(u'utf_16_le')
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'ProfileGuid', data=value_data,
        data_type=dfwinreg_definitions.REG_SZ)
    unmanaged_subkey.AddValue(registry_value)

    value_data = b'\x00\x00\x00\x08'
    registry_value = dfwinreg_fake.FakeWinRegistryValue(
        u'Source', data=value_data,
        data_type=dfwinreg_definitions.REG_DWORD_BIG_ENDIAN)
    unmanaged_subkey.AddValue(registry_value)

    return registry_key

  def testProcess(self):
    """Tests the Process function on created key."""
    key_path = (
        u'HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows NT\\CurrentVersion')
    time_string = u'2013-01-30 10:47:57'
    registry_key = self._CreateTestKey(key_path, time_string)

    plugin = networks.NetworksPlugin()
    storage_writer = self._ParseKeyWithPlugin(registry_key, plugin)

    self.assertEqual(storage_writer.number_of_events, 4)

    events = list(storage_writer.GetSortedEvents())

    event = events[0]

    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2014-05-06 17:02:19.795')
    self.assertEqual(event.timestamp, expected_timestamp)
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_CREATION)

    expected_message = (
        u'SSID: Network '
        u'Description: Network '
        u'Connection Type: Wired '
        u'Default Gateway Mac: 00:50:56:ea:6c:ec '
        u'DNS Suffix: localdomain')
    expected_short_message = u'{0:s}...'.format(expected_message[:77])

    self._TestGetMessageStrings(event, expected_message, expected_short_message)

    event = events[3]

    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2015-01-27 15:15:27.965')
    self.assertEqual(event.timestamp, expected_timestamp)
    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_LAST_CONNECTED)

    expected_message = (
        u'SSID: My Awesome Wifi Hotspot '
        u'Description: My Awesome Wifi Hotspot '
        u'Connection Type: Wireless')
    expected_short_message = u'{0:s}...'.format(expected_message[:77])

    self._TestGetMessageStrings(event, expected_message, expected_short_message)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The 4n6time SQLite database output module CLI arguments helper."""

from plaso.lib import errors
from plaso.cli.helpers import interface
from plaso.cli.helpers import shared_4n6time_output
from plaso.cli.helpers import manager
from plaso.output import sqlite_4n6time


class SQLite4n6TimeOutputArgumentsHelper(interface.ArgumentsHelper):
  """4n6time SQLite database output module CLI arguments helper."""

  NAME = u'4n6time_sqlite'
  CATEGORY = u'output'
  DESCRIPTION = u'Argument helper for the 4n6Time SQLite output module.'

  @classmethod
  def AddArguments(cls, argument_group):
    """Adds command line arguments the helper supports to an argument group.

    This function takes an argument parser or an argument group object and adds
    to it all the command line arguments this helper supports.

    Args:
      argument_group (argparse._ArgumentGroup|argparse.ArgumentParser):
          argparse group.
    """
    shared_4n6time_output.Shared4n6TimeOutputArgumentsHelper.AddArguments(
        argument_group)

  @classmethod
  def ParseOptions(cls, options, output_module):
    """Parses and validates options.

    Args:
      options (argparse.Namespace): parser options.
      output_module (OutputModule): output module to configure.

    Raises:
      BadConfigObject: when the output module object is of the wrong type.
      BadConfigOption: when the output filename was not provided.
    """
    if not isinstance(output_module, sqlite_4n6time.SQLite4n6TimeOutputModule):
      raise errors.BadConfigObject(
          u'Output module is not an instance of SQLite4n6TimeOutputModule')

    shared_4n6time_output.Shared4n6TimeOutputArgumentsHelper.ParseOptions(
        options, output_module)

    filename = getattr(options, u'write', None)
    if not filename:
      raise errors.BadConfigOption(
          u'Output filename was not provided use "-w filename" to specify.')

    output_module.SetFilename(filename)


manager.ArgumentHelperManager.RegisterHelper(SQLite4n6TimeOutputArgumentsHelper)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Tests for the MSIE Zone settings Windows Registry plugin."""

import unittest

from plaso.formatters import winreg  # pylint: disable=unused-import
from plaso.lib import timelib
from plaso.parsers.winreg_plugins import msie_zones

from tests import test_lib as shared_test_lib
from tests.parsers.winreg_plugins import test_lib


class MsieZoneSettingsPluginTest(test_lib.RegistryPluginTestCase):
  """Tests for Internet Settings Zones plugin."""

  @shared_test_lib.skipUnlessHasTestFile([u'NTUSER-WIN7.DAT'])
  def testProcessNtuserLockdownZones(self):
    """Tests the Process function on a Lockdown_Zones key."""
    test_file_entry = self._GetTestFileEntry([u'NTUSER-WIN7.DAT'])
    key_path = (
        u'HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\'
        u'Internet Settings\\Lockdown_Zones')

    win_registry = self._GetWinRegistryFromFileEntry(test_file_entry)
    registry_key = win_registry.GetKeyByPath(key_path)

    plugin = msie_zones.MsieZoneSettingsPlugin()
    storage_writer = self._ParseKeyWithPlugin(
        registry_key, plugin, file_entry=test_file_entry)

    self.assertEqual(storage_writer.number_of_events, 6)

    events = list(storage_writer.GetEvents())

    event = events[1]

    # This should just be the plugin name, as we're invoking it directly,
    # and not through the parser.
    self.assertEqual(event.parser, plugin.plugin_name)

    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2011-09-16 21:12:40.145514')
    self.assertEqual(event.timestamp, expected_timestamp)

    regvalue_identifier = u'[1200] Run ActiveX controls and plug-ins'
    expected_value = u'3 (Not Allowed)'
    self._TestRegvalue(event, regvalue_identifier, expected_value)

    expected_message = (
        u'[{0:s}\\0 (My Computer)] '
        u'[1200] Run ActiveX controls and plug-ins: 3 (Not Allowed) '
        u'[1400] Active scripting: 1 (Prompt User) '
        u'[CurrentLevel]: 0 '
        u'[Description]: Your computer '
        u'[DisplayName]: Computer '
        u'[Flags]: 33 '
        u'[Icon]: shell32.dll#0016 '
        u'[LowIcon]: inetcpl.cpl#005422 '
        u'[PMDisplayName]: Computer '
        u'[Protected Mode]').format(key_path)
    expected_short_message = u'{0:s}...'.format(expected_message[:77])

    self._TestGetMessageStrings(
        event, expected_message, expected_short_message)

  @shared_test_lib.skipUnlessHasTestFile([u'NTUSER-WIN7.DAT'])
  def testProcessNtuserZones(self):
    """Tests the Process function on a Zones key."""
    test_file_entry = self._GetTestFileEntry([u'NTUSER-WIN7.DAT'])
    key_path = (
        u'HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\'
        u'Internet Settings\\Zones')

    win_registry = self._GetWinRegistryFromFileEntry(test_file_entry)
    registry_key = win_registry.GetKeyByPath(key_path)

    plugin = msie_zones.MsieZoneSettingsPlugin()
    storage_writer = self._ParseKeyWithPlugin(
        registry_key, plugin, file_entry=test_file_entry)

    self.assertEqual(storage_writer.number_of_events, 6)

    events = list(storage_writer.GetEvents())

    event = events[1]

    # This should just be the plugin name, as we're invoking it directly,
    # and not through the parser.
    self.assertEqual(event.parser, plugin.plugin_name)

    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2011-09-16 21:12:40.145514')
    self.assertEqual(event.timestamp, expected_timestamp)

    regvalue_identifier = u'[1200] Run ActiveX controls and plug-ins'
    expected_value = u'0 (Allow)'
    self._TestRegvalue(event, regvalue_identifier, expected_value)

    expected_message = (
        u'[{0:s}\\0 (My Computer)] '
        u'[1200] Run ActiveX controls and plug-ins: 0 (Allow) '
        u'[1400] Active scripting: 0 (Allow) '
        u'[2001] .NET: Run components signed with Authenticode: 3 (Not '
        u'Allowed) '
        u'[2004] .NET: Run components not signed with Authenticode: 3 (Not '
        u'Allowed) '
        u'[2007] UNKNOWN: 3 '
        u'[CurrentLevel]: 0 '
        u'[Description]: Your computer '
        u'[DisplayName]: Computer '
        u'[Flags]: 33 [Icon]: shell32.dll#0016 '
        u'[LowIcon]: inetcpl.cpl#005422 '
        u'[PMDisplayName]: Computer '
        u'[Protected Mode]').format(key_path)
    expected_short_message = u'{0:s}...'.format(expected_message[:77])

    self._TestGetMessageStrings(
        event, expected_message, expected_short_message)

  @shared_test_lib.skipUnlessHasTestFile([u'SOFTWARE'])
  def testProcessSoftwareLockdownZones(self):
    """Tests the Process function on a Lockdown_Zones key."""
    test_file_entry = self._GetTestFileEntry([u'SOFTWARE'])
    key_path = (
        u'HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\'
        u'Internet Settings\\Lockdown_Zones')

    win_registry = self._GetWinRegistryFromFileEntry(test_file_entry)
    registry_key = win_registry.GetKeyByPath(key_path)

    plugin = msie_zones.MsieZoneSettingsPlugin()
    storage_writer = self._ParseKeyWithPlugin(
        registry_key, plugin, file_entry=test_file_entry)

    self.assertEqual(storage_writer.number_of_events, 6)

    events = list(storage_writer.GetEvents())

    event = events[1]

    # This should just be the plugin name, as we're invoking it directly,
    # and not through the parser.
    self.assertEqual(event.parser, plugin.plugin_name)

    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2011-08-28 21:32:44.937675')
    self.assertEqual(event.timestamp, expected_timestamp)

    regvalue_identifier = u'[1200] Run ActiveX controls and plug-ins'
    expected_value = u'3 (Not Allowed)'
    self._TestRegvalue(event, regvalue_identifier, expected_value)

    expected_message = (
        u'[{0:s}\\0 (My Computer)] '
        u'[1001] Download signed ActiveX controls: 1 (Prompt User) '
        u'[1004] Download unsigned ActiveX controls: 3 (Not Allowed) '
        u'[1200] Run ActiveX controls and plug-ins: 3 (Not Allowed) '
        u'[1201] Initialize and script ActiveX controls not marked as safe: 3 '
        u'(Not Allowed) '
        u'[1206] Allow scripting of IE Web browser control: 0 '
        u'[1207] Reserved: 3 '
        u'[1208] Allow previously unused ActiveX controls to run without '
        u'prompt: 3 '
        u'[1209] Allow Scriptlets: 3 '
        u'[120A] Override Per-Site (domain-based) ActiveX restrictions: 3 '
        u'[120B] Override Per-Site (domain-based) ActiveX restrictions: 0 '
        u'[1400] Active scripting: 1 (Prompt User) '
        u'[1402] Scripting of Java applets: 0 (Allow) '
        u'[1405] Script ActiveX controls marked as safe for scripting: 0 '
        u'(Allow) '
        u'[1406] Access data sources across domains: 0 (Allow) '
        u'[1407] Allow Programmatic clipboard access: 1 (Prompt User) '
        u'[1408] Reserved: 3 '
        u'[1409] UNKNOWN: 3 '
        u'[1601] Submit non-encrypted form data: 0 (Allow) '
        u'[1604] Font download: 0 (Allow) '
        u'[1605] Run Java: 0 '
        u'[1606] Userdata persistence: 0 (Allow) '
        u'[1607] Navigate sub-frames across different domains: 0 (Allow) '
        u'[1608] Allow META REFRESH: 0 (Allow) '
        u'[1609] Display mixed content: 1 (Prompt User) '
        u'[160A] Include local directory path when uploading files to a '
        u'server: 0 '
        u'[1802] Drag and drop or copy and paste files: 0 (Allow) '
        u'[1803] File Download: 0 (Allow) '
        u'[1804] Launching programs and files in an IFRAME: 0 (Allow) '
        u'[1805] Launching programs and files in webview: 0 '
        u'[1806] Launching applications and unsafe files: 0 '
        u'[1807] Reserved: 0 '
        u'[1808] Reserved: 0 '
        u'[1809] Use Pop-up Blocker: 3 (Not Allowed) '
        u'[180A] Reserved: 0 '
        u'[180C] Reserved: 0 '
        u'[180D] Reserved: 0 '
        u'[180E] UNKNOWN: 0 '
        u'[180F] UNKNOWN: 0 '
        u'[1A00] User Authentication: Logon: 0x00000000 (Automatic logon with '
        u'current user name and password) '
        u'[1A02] Allow persistent cookies that are stored on your computer: 0 '
        u'[1A03] Allow per-session cookies (not stored): 0 '
        u'[1A04] Don\'t prompt for client cert selection when no certs exists: '
        u'3 (Not Allowed) '
        u'[1A05] Allow 3rd party persistent cookies: 0 '
        u'[1A06] Allow 3rd party session cookies: 0 '
        u'[1A10] Privacy Settings: 0 '
        u'[1C00] Java permissions: 0x00000000 (Disable Java) '
        u'[2000] Binary and script behaviors: 0x00010000 '
        u'(Administrator approved) '
        u'[2005] UNKNOWN: 3 '
        u'[2100] Open files based on content, not file extension: 3 '
        u'(Not Allowed) '
        u'[2101] Web sites in less privileged zone can navigate into this '
        u'zone: 3 (Not Allowed) '
        u'[2102] Allow script initiated windows without size/position '
        u'constraints: '
        u'3 (Not Allowed) '
        u'[2103] Allow status bar updates via script: 3 '
        u'[2104] Allow websites to open windows without address or status '
        u'bars: 3 '
        u'[2105] Allow websites to prompt for information using scripted '
        u'windows: 3 '
        u'[2106] UNKNOWN: 3 '
        u'[2107] UNKNOWN: 3 '
        u'[2200] Automatic prompting for file downloads: 3 (Not Allowed) '
        u'[2201] Automatic prompting for ActiveX controls: 3 (Not Allowed) '
        u'[2301] Use Phishing Filter: 3 '
        u'[2400] .NET: XAML browser applications: 0 '
        u'[2401] .NET: XPS documents: 0 '
        u'[2402] .NET: Loose XAML: 0 '
        u'[2500] Turn on Protected Mode: 3 '
        u'[2600] Enable .NET Framework setup: 0 '
        u'[2700] UNKNOWN: 3 '
        u'[2701] UNKNOWN: 3 '
        u'[2702] UNKNOWN: 3 '
        u'[2703] UNKNOWN: 3 '
        u'[2708] UNKNOWN: 0 '
        u'[2709] UNKNOWN: 0 '
        u'[CurrentLevel]: 0 '
        u'[Description]: Your computer '
        u'[DisplayName]: Computer '
        u'[Flags]: 33 '
        u'[Icon]: shell32.dll#0016 '
        u'[LowIcon]: inetcpl.cpl#005422 '
        u'[PMDisplayName]: Computer '
        u'[Protected Mode]').format(key_path)
    expected_short_message = u'{0:s}...'.format(expected_message[:77])

    self._TestGetMessageStrings(
        event, expected_message, expected_short_message)

  @shared_test_lib.skipUnlessHasTestFile([u'SOFTWARE'])
  def testProcessSoftwareZones(self):
    """Tests the Process function on a Zones key."""
    test_file_entry = self._GetTestFileEntry([u'SOFTWARE'])
    key_path = (
        u'HKEY_LOCAL_MACHINE\\Software\\Microsoft\\Windows\\CurrentVersion\\'
        u'Internet Settings\\Zones')

    win_registry = self._GetWinRegistryFromFileEntry(test_file_entry)
    registry_key = win_registry.GetKeyByPath(key_path)

    plugin = msie_zones.MsieZoneSettingsPlugin()
    storage_writer = self._ParseKeyWithPlugin(
        registry_key, plugin, file_entry=test_file_entry)

    self.assertEqual(storage_writer.number_of_events, 6)

    events = list(storage_writer.GetEvents())

    event = events[1]

    # This should just be the plugin name, as we're invoking it directly,
    # and not through the parser.
    self.assertEqual(event.parser, plugin.plugin_name)

    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2011-08-28 21:32:44.937675')
    self.assertEqual(event.timestamp, expected_timestamp)

    regvalue_identifier = u'[1200] Run ActiveX controls and plug-ins'
    expected_value = u'0 (Allow)'
    self._TestRegvalue(event, regvalue_identifier, expected_value)

    expected_message = (
        u'[{0:s}\\0 (My Computer)] '
        u'[1001] Download signed ActiveX controls: 0 (Allow) '
        u'[1004] Download unsigned ActiveX controls: 0 (Allow) '
        u'[1200] Run ActiveX controls and plug-ins: 0 (Allow) '
        u'[1201] Initialize and script ActiveX controls not marked as safe: 1 '
        u'(Prompt User) '
        u'[1206] Allow scripting of IE Web browser control: 0 '
        u'[1207] Reserved: 0 '
        u'[1208] Allow previously unused ActiveX controls to run without '
        u'prompt: 0 '
        u'[1209] Allow Scriptlets: 0 '
        u'[120A] Override Per-Site (domain-based) ActiveX restrictions: 0 '
        u'[120B] Override Per-Site (domain-based) ActiveX restrictions: 0 '
        u'[1400] Active scripting: 0 (Allow) '
        u'[1402] Scripting of Java applets: 0 (Allow) '
        u'[1405] Script ActiveX controls marked as safe for scripting: 0 '
        u'(Allow) '
        u'[1406] Access data sources across domains: 0 (Allow) '
        u'[1407] Allow Programmatic clipboard access: 0 (Allow) '
        u'[1408] Reserved: 0 '
        u'[1409] UNKNOWN: 3 '
        u'[1601] Submit non-encrypted form data: 0 (Allow) '
        u'[1604] Font download: 0 (Allow) '
        u'[1605] Run Java: 0 '
        u'[1606] Userdata persistence: 0 (Allow) '
        u'[1607] Navigate sub-frames across different domains: 0 (Allow) '
        u'[1608] Allow META REFRESH: 0 (Allow) '
        u'[1609] Display mixed content: 1 (Prompt User) '
        u'[160A] Include local directory path when uploading files to a '
        u'server: 0 '
        u'[1802] Drag and drop or copy and paste files: 0 (Allow) '
        u'[1803] File Download: 0 (Allow) '
        u'[1804] Launching programs and files in an IFRAME: 0 (Allow) '
        u'[1805] Launching programs and files in webview: 0 '
        u'[1806] Launching applications and unsafe files: 0 '
        u'[1807] Reserved: 0 '
        u'[1808] Reserved: 0 '
        u'[1809] Use Pop-up Blocker: 3 (Not Allowed) '
        u'[180A] Reserved: 0 '
        u'[180C] Reserved: 0 '
        u'[180D] Reserved: 0 '
        u'[180E] UNKNOWN: 0 '
        u'[180F] UNKNOWN: 0 '
        u'[1A00] User Authentication: Logon: 0x00000000 (Automatic logon with '
        u'current user name and password) '
        u'[1A02] Allow persistent cookies that are stored on your computer: 0 '
        u'[1A03] Allow per-session cookies (not stored): 0 '
        u'[1A04] Don\'t prompt for client cert selection when no certs exists: '
        u'0 (Allow) '
        u'[1A05] Allow 3rd party persistent cookies: 0 '
        u'[1A06] Allow 3rd party session cookies: 0 '
        u'[1A10] Privacy Settings: 0 '
        u'[1C00] Java permissions: 0x00020000 (Medium safety) '
        u'[2000] Binary and script behaviors: 0 (Allow) '
        u'[2001] .NET: Run components signed with Authenticode: '
        u'3 (Not Allowed) '
        u'[2004] .NET: Run components not signed with Authenticode: '
        u'3 (Not Allowed) '
        u'[2005] UNKNOWN: 0 '
        u'[2007] UNKNOWN: 3 '
        u'[2100] Open files based on content, not file extension: 0 (Allow) '
        u'[2101] Web sites in less privileged zone can navigate into this '
        u'zone: 3 (Not Allowed) '
        u'[2102] Allow script initiated windows without size/position '
        u'constraints: 0 (Allow) '
        u'[2103] Allow status bar updates via script: 0 '
        u'[2104] Allow websites to open windows without address or status '
        u'bars: 0 '
        u'[2105] Allow websites to prompt for information using scripted '
        u'windows: 0 '
        u'[2106] UNKNOWN: 0 '
        u'[2107] UNKNOWN: 0 '
        u'[2200] Automatic prompting for file downloads: 0 (Allow) '
        u'[2201] Automatic prompting for ActiveX controls: 0 (Allow) '
        u'[2300] Allow web pages to use restricted protocols for active '
        u'content: 1 (Prompt User) '
        u'[2301] Use Phishing Filter: 3 '
        u'[2400] .NET: XAML browser applications: 0 '
        u'[2401] .NET: XPS documents: 0 '
        u'[2402] .NET: Loose XAML: 0 '
        u'[2500] Turn on Protected Mode: 3 '
        u'[2600] Enable .NET Framework setup: 0 '
        u'[2700] UNKNOWN: 3 '
        u'[2701] UNKNOWN: 0 '
        u'[2702] UNKNOWN: 3 '
        u'[2703] UNKNOWN: 3 '
        u'[2708] UNKNOWN: 0 '
        u'[2709] UNKNOWN: 0 '
        u'[CurrentLevel]: 0 '
        u'[Description]: Your computer '
        u'[DisplayName]: Computer '
        u'[Flags]: 33 '
        u'[Icon]: shell32.dll#0016 '
        u'[LowIcon]: inetcpl.cpl#005422 '
        u'[PMDisplayName]: Computer '
        u'[Protected Mode]').format(key_path)
    expected_short_message = u'{0:s}...'.format(expected_message[:77])

    self._TestGetMessageStrings(
        event, expected_message, expected_short_message)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""This file contains a appfirewall.log (Mac OS X Firewall) parser."""

import logging

import pyparsing

from dfdatetime import time_elements as dfdatetime_time_elements

from plaso.containers import events
from plaso.containers import time_events
from plaso.lib import errors
from plaso.lib import definitions
from plaso.lib import timelib
from plaso.parsers import manager
from plaso.parsers import text_parser


__author__ = 'Joaquin Moreno Garijo (Joaquin.MorenoGarijo.2013@live.rhul.ac.uk)'


class MacAppFirewallLogEventData(events.EventData):
  """Mac OS X Firewall log event data.

  Attributes:
    action (str): action.
    agent (str): agent that save the log.
    computer_name (str): name of the computer.
    process_name (str): name of the entity that tried do the action.
    status (str): saved status action.
  """

  DATA_TYPE = u'mac:appfirewall:line'

  def __init__(self):
    """Initializes event data."""
    super(MacAppFirewallLogEventData, self).__init__(data_type=self.DATA_TYPE)
    self.action = None
    self.agent = None
    self.computer_name = None
    self.process_name = None
    self.status = None


class MacAppFirewallParser(text_parser.PyparsingSingleLineTextParser):
  """Parse text based on appfirewall.log file."""

  NAME = u'mac_appfirewall_log'
  DESCRIPTION = u'Parser for appfirewall.log files.'

  ENCODING = u'utf-8'

  # Define how a log line should look like.
  # Example: 'Nov  2 04:07:35 DarkTemplar-2.local socketfilterfw[112] '
  #          '<Info>: Dropbox: Allow (in:0 out:2)'
  # INFO: process_name is going to have a white space at the beginning.

  DATE_TIME = pyparsing.Group(
      text_parser.PyparsingConstants.THREE_LETTERS.setResultsName(u'month') +
      text_parser.PyparsingConstants.ONE_OR_TWO_DIGITS.setResultsName(u'day') +
      text_parser.PyparsingConstants.TIME_ELEMENTS)

  FIREWALL_LINE = (
      DATE_TIME.setResultsName(u'date_time') +
      pyparsing.Word(pyparsing.printables).setResultsName(u'computer_name') +
      pyparsing.Word(pyparsing.printables).setResultsName(u'agent') +
      pyparsing.Literal(u'<').suppress() +
      pyparsing.CharsNotIn(u'>').setResultsName(u'status') +
      pyparsing.Literal(u'>:').suppress() +
      pyparsing.CharsNotIn(u':').setResultsName(u'process_name') +
      pyparsing.Literal(u':') +
      pyparsing.SkipTo(pyparsing.lineEnd).setResultsName(u'action'))

  # Repeated line.
  # Example: Nov 29 22:18:29 --- last message repeated 1 time ---

  REPEATED_LINE = (
      DATE_TIME.setResultsName(u'date_time') +
      pyparsing.Literal(u'---').suppress() +
      pyparsing.CharsNotIn(u'---').setResultsName(u'process_name') +
      pyparsing.Literal(u'---').suppress())

  LINE_STRUCTURES = [
      (u'logline', FIREWALL_LINE),
      (u'repeated', REPEATED_LINE)]

  def __init__(self):
    """Initializes a parser object."""
    super(MacAppFirewallParser, self).__init__()
    self._last_month = 0
    self._previous_structure = None
    self._year_use = 0

  def _GetTimeElementsTuple(self, structure):
    """Retrieves a time elements tuple from the structure.

    Args:
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.

    Returns:
      tuple: contains:
        year (int): year.
        month (int): month, where 1 represents January.
        day_of_month (int): day of month, where 1 is the first day of the month.
        hours (int): hours.
        minutes (int): minutes.
        seconds (int): seconds.
    """
    month, day, hours, minutes, seconds = structure.date_time

    # Note that dfdatetime_time_elements.TimeElements will raise ValueError
    # for an invalid month.
    month = timelib.MONTH_DICT.get(month.lower(), 0)

    if month != 0 and month < self._last_month:
      # Gap detected between years.
      self._year_use += 1

    return (self._year_use, month, day, hours, minutes, seconds)

  def _ParseLogLine(self, parser_mediator, structure, key):
    """Parse a single log line and produce an event object.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      key (str): identifier of the structure of tokens.
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.
    """
    time_elements_tuple = self._GetTimeElementsTuple(structure)

    try:
      date_time = dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)
    except ValueError:
      parser_mediator.ProduceExtractionError(
          u'invalid date time value: {0!s}'.format(structure.date_time))
      return

    self._last_month = time_elements_tuple[1]

    # If the actual entry is a repeated entry, we take the basic information
    # from the previous entry, but using the timestmap from the actual entry.
    if key == u'logline':
      self._previous_structure = structure
    else:
      structure = self._previous_structure

    # Pyparsing reads in RAW, but the text is in UTF8.
    try:
      action = structure.action.decode(u'utf-8')
    except UnicodeDecodeError:
      logging.warning(
          u'Decode UTF8 failed, the message string may be cut short.')
      action = structure.action.decode(u'utf-8', u'ignore')

    event_data = MacAppFirewallLogEventData()
    event_data.action = action
    event_data.agent = structure.agent
    event_data.computer_name = structure.computer_name
    # Due to the use of CharsNotIn pyparsing structure contains whitespaces
    # that need to be removed.
    event_data.process_name = structure.process_name.strip()
    event_data.status = structure.status

    event = time_events.DateTimeValuesEvent(
        date_time, definitions.TIME_DESCRIPTION_ADDED)
    parser_mediator.ProduceEventWithEventData(event, event_data)

  def ParseRecord(self, parser_mediator, key, structure):
    """Parses a log record structure and produces events.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      key (str): identifier of the structure of tokens.
      structure (pyparsing.ParseResults): structure of tokens derived from
          a line of a text file.

    Raises:
      ParseError: when the structure type is unknown.
    """
    if key not in (u'logline', u'repeated'):
      raise errors.ParseError(
          u'Unable to parse record, unknown structure: {0:s}'.format(key))

    self._ParseLogLine(parser_mediator, structure, key)

  def VerifyStructure(self, parser_mediator, line):
    """Verify that this file is a Mac AppFirewall log file.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      line (bytes): line from a text file.

    Returns:
      bool: True if the line is in the expected format, False if not.
    """
    self._last_month = 0
    self._year_use = parser_mediator.GetEstimatedYear()

    try:
      structure = self.FIREWALL_LINE.parseString(line)
    except pyparsing.ParseException as exception:
      logging.debug((
          u'Unable to parse file as a Mac AppFirewall log file with error: '
          u'{0:s}').format(exception))
      return False

    if structure.action != u'creating /var/log/appfirewall.log':
      logging.debug(
          u'Not a Mac AppFirewall log file, invalid action: {0!s}'.format(
              structure.action))
      return False

    if structure.status != u'Error':
      logging.debug(
          u'Not a Mac AppFirewall log file, invalid status: {0!s}'.format(
              structure.status))
      return False

    time_elements_tuple = self._GetTimeElementsTuple(structure)

    try:
      dfdatetime_time_elements.TimeElements(
          time_elements_tuple=time_elements_tuple)
    except ValueError:
      logging.debug((
          u'Not a Mac AppFirewall log file, invalid date and time: '
          u'{0!s}').format(structure.date_time))
      return False

    self._last_month = time_elements_tuple[1]

    return True


manager.ParsersManager.RegisterParser(MacAppFirewallParser)



# -*- coding: utf-8 -*-
"""Bencode parser plugin for Transmission BitTorrent files."""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.containers import time_events
from plaso.lib import definitions
from plaso.parsers import bencode_parser
from plaso.parsers.bencode_plugins import interface


class TransmissionEventData(events.EventData):
  """Transmission BitTorrent event data.

  Attributes:
    destination (str): downloaded file name within .torrent file
    seedtime (int): number of seconds client seeded torrent
  """

  DATA_TYPE = u'p2p:bittorrent:transmission'

  def __init__(self):
    """Initializes event data."""
    super(TransmissionEventData, self).__init__(data_type=self.DATA_TYPE)
    self.destination = None
    self.seedtime = None


class TransmissionPlugin(interface.BencodePlugin):
  """Parse Transmission BitTorrent activity file for current torrents."""

  NAME = u'bencode_transmission'
  DESCRIPTION = u'Parser for Transmission bencoded files.'

  BENCODE_KEYS = frozenset([
      u'activity-date', u'done-date', u'added-date', u'destination',
      u'seeding-time-seconds'])

  def GetEntries(self, parser_mediator, data=None, **unused_kwargs):
    """Extract data from Transmission's resume folder files.

    This is the main parsing engine for the parser. It determines if
    the selected file is the proper file to parse and extracts current
    running torrents.

    Transmission stores an individual Bencoded file for each active download
    in a folder named resume under the user's application data folder.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      data (Optional[dict[str, object]]): bencode data values.
    """
    seeding_time = data.get(u'seeding-time-seconds', None)

    event_data = TransmissionEventData()
    event_data.destination = data.get(u'destination', None)
    # Convert seconds to minutes.
    event_data.seedtime, _ = divmod(seeding_time, 60)

    # Create timeline events based on extracted values.
    timestamp = data.get(u'added-date', None)
    if timestamp:
      date_time = dfdatetime_posix_time.PosixTime(timestamp=timestamp)
      event = time_events.DateTimeValuesEvent(
          date_time, definitions.TIME_DESCRIPTION_ADDED)
      parser_mediator.ProduceEventWithEventData(event, event_data)

    timestamp = data.get(u'done-date', None)
    if timestamp:
      date_time = dfdatetime_posix_time.PosixTime(timestamp=timestamp)
      event = time_events.DateTimeValuesEvent(
          date_time, definitions.TIME_DESCRIPTION_FILE_DOWNLOADED)
      parser_mediator.ProduceEventWithEventData(event, event_data)

    timestamp = data.get(u'activity-date', None)
    if timestamp:
      date_time = dfdatetime_posix_time.PosixTime(timestamp=timestamp)
      event = time_events.DateTimeValuesEvent(
          date_time, definitions.TIME_DESCRIPTION_LAST_ACCESS)
      parser_mediator.ProduceEventWithEventData(event, event_data)


bencode_parser.BencodeParser.RegisterPlugin(TransmissionPlugin)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Tests for the extraction CLI arguments helper."""

import argparse
import unittest

from plaso.cli import tools
from plaso.cli.helpers import extraction
from plaso.lib import errors

from tests.cli import test_lib as cli_test_lib


class ExtractionArgumentsHelperTest(cli_test_lib.CLIToolTestCase):
  """Tests for the extraction CLI arguments helper."""

  # pylint: disable=protected-access

  _EXPECTED_OUTPUT = u'\n'.join([
      u'usage: cli_helper.py [--preferred_year YEAR] [-p] [--process_archives]',
      u'                     [--skip_compressed_streams]',
      u'',
      u'Test argument parser.',
      u'',
      u'optional arguments:',
      u'  --preferred_year YEAR, --preferred-year YEAR',
      (u'                        When a format\'s timestamp does not include '
       u'a year,'),
      (u'                        e.g. syslog, use this as the initial year '
       u'instead of'),
      u'                        attempting auto-detection.',
      u'  --process_archives, --process-archives',
      (u'                        Process file entries embedded within archive '
       u'files,'),
      (u'                        such as archive.tar and archive.zip. This '
       u'can make'),
      u'                        processing significantly slower.',
      u'  --skip_compressed_streams, --skip-compressed-streams',
      u'                        Skip processing file content within compressed',
      u'                        streams, such as syslog.gz and syslog.bz2.',
      (u'  -p, --preprocess      Turn on preprocessing. Preprocessing is '
       u'turned on by'),
      (u'                        default when parsing image files, however if '
       u'a mount'),
      (u'                        point is being parsed then this parameter '
       u'needs to be'),
      u'                        set manually.',
      u''])

  def testAddArguments(self):
    """Tests the AddArguments function."""
    argument_parser = argparse.ArgumentParser(
        prog=u'cli_helper.py', description=u'Test argument parser.',
        add_help=False,
        formatter_class=cli_test_lib.SortedArgumentsHelpFormatter)

    extraction.ExtractionArgumentsHelper.AddArguments(argument_parser)

    output = self._RunArgparseFormatHelp(argument_parser)
    self.assertEqual(output, self._EXPECTED_OUTPUT)

  def testParseOptions(self):
    """Tests the ParseOptions function."""
    options = cli_test_lib.TestOptions()

    test_tool = tools.CLITool()
    extraction.ExtractionArgumentsHelper.ParseOptions(options, test_tool)

    self.assertFalse(test_tool._force_preprocessing)
    self.assertIsNone(test_tool._preferred_year)
    self.assertFalse(test_tool._process_archives)
    self.assertTrue(test_tool._process_compressed_streams)

    with self.assertRaises(errors.BadConfigObject):
      extraction.ExtractionArgumentsHelper.ParseOptions(options, None)

    # TODO: improve test coverage.


if __name__ == '__main__':
  unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Tests for the Explorer ProgramsCache Windows Registry plugin."""

import unittest

from plaso.formatters import winreg  # pylint: disable=unused-import
from plaso.lib import definitions
from plaso.lib import timelib
from plaso.parsers.winreg_plugins import programscache

from tests import test_lib as shared_test_lib
from tests.parsers.winreg_plugins import test_lib


class ExplorerProgramCachePluginTest(test_lib.RegistryPluginTestCase):
  """Tests for the Explorer ProgramsCache Windows Registry plugin."""

  @shared_test_lib.skipUnlessHasTestFile([u'NTUSER.DAT'])
  def testProcessStartPage(self):
    """Tests the Process function on a StartPage key."""
    test_file_entry = self._GetTestFileEntry([u'NTUSER.DAT'])
    key_path = (
        u'HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\'
        u'Explorer\\StartPage')

    win_registry = self._GetWinRegistryFromFileEntry(test_file_entry)
    registry_key = win_registry.GetKeyByPath(key_path)

    plugin = programscache.ExplorerProgramsCachePlugin()
    storage_writer = self._ParseKeyWithPlugin(
        registry_key, plugin, file_entry=test_file_entry)

    self.assertEqual(storage_writer.number_of_events, 77)

    events = list(storage_writer.GetEvents())

    # The ProgramsCache entry shell item event.
    event = events[0]

    expected_parser = u'explorer_programscache/shell_items'
    self.assertEqual(event.parser, expected_parser)

    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_CREATION)
    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2009-08-04 15:12:24')
    self.assertEqual(event.timestamp, expected_timestamp)

    expected_data_type = u'windows:shell_item:file_entry'
    self.assertEqual(event.data_type, expected_data_type)

    expected_message = (
        u'Name: Programs '
        u'Long name: Programs '
        u'Localized name: @shell32.dll,-21782 '
        u'Shell item path: Programs '
        u'Origin: {0:s} ProgramsCache').format(key_path)
    expected_short_message = (
        u'Name: Programs '
        u'Origin: HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\'
        u'CurrentVe...')

    self._TestGetMessageStrings(event, expected_message, expected_short_message)

    # The ProgramsCache list event.
    event = events[75]

    expected_parser = u'explorer_programscache'
    self.assertEqual(event.parser, expected_parser)

    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_WRITTEN)
    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2009-08-04 15:22:18.419625')
    self.assertEqual(event.timestamp, expected_timestamp)

    expected_data_type = u'windows:registry:list'
    self.assertEqual(event.data_type, expected_data_type)

    expected_message = (
        u'Key: {0:s} '
        u'Value: ProgramsCache '
        u'List: ProgramsCache ['
        u'0: Programs '
        u'1: Internet Explorer.lnk '
        u'2: Outlook Express.lnk '
        u'3: Remote Assistance.lnk '
        u'4: Windows Media Player.lnk '
        u'5: Programs Accessories '
        u'6: Address Book.lnk '
        u'7: Command Prompt.lnk '
        u'8: Notepad.lnk '
        u'9: Program Compatibility Wizard.lnk '
        u'10: Synchronize.lnk '
        u'11: Tour Windows XP.lnk '
        u'12: Windows Explorer.lnk '
        u'13: Programs Accessories\\Accessibility '
        u'14: Magnifier.lnk '
        u'15: Narrator.lnk '
        u'16: On-Screen Keyboard.lnk '
        u'17: Utility Manager.lnk '
        u'18: Programs Accessories\\System Tools '
        u'19: Internet Explorer (No Add-ons).lnk]').format(key_path)
    expected_short_message = u'{0:s}...'.format(expected_message[:77])

    self._TestGetMessageStrings(event, expected_message, expected_short_message)

    # The Windows Registry key event.
    event = events[76]

    expected_parser = u'explorer_programscache'
    self.assertEqual(event.parser, expected_parser)

    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_WRITTEN)
    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2009-08-04 15:22:18.419625')
    self.assertEqual(event.timestamp, expected_timestamp)

    expected_data_type = u'windows:registry:key_value'
    self.assertEqual(event.data_type, expected_data_type)

  @shared_test_lib.skipUnlessHasTestFile([u'NTUSER-WIN7.DAT'])
  def testProcessStartPage2(self):
    """Tests the Process function on a StartPage2 key."""
    test_file_entry = self._GetTestFileEntry([u'NTUSER-WIN7.DAT'])
    key_path = (
        u'HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\'
        u'Explorer\\StartPage2')

    win_registry = self._GetWinRegistryFromFileEntry(test_file_entry)
    registry_key = win_registry.GetKeyByPath(key_path)

    plugin = programscache.ExplorerProgramsCachePlugin()
    storage_writer = self._ParseKeyWithPlugin(
        registry_key, plugin, file_entry=test_file_entry)

    self.assertEqual(storage_writer.number_of_events, 118)

    events = list(storage_writer.GetEvents())

    event = events[0]

    expected_parser = u'explorer_programscache/shell_items'
    self.assertEqual(event.parser, expected_parser)

    self.assertEqual(
        event.timestamp_desc, definitions.TIME_DESCRIPTION_CREATION)
    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2010-11-10 07:50:38')
    self.assertEqual(event.timestamp, expected_timestamp)

    expected_data_type = u'windows:shell_item:file_entry'
    self.assertEqual(event.data_type, expected_data_type)


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""A plugin to tag events according to rules in a tag file."""

import logging
import re
import os

from efilter import ast as efilter_ast
from efilter import api as efilter_api
from efilter import errors as efilter_errors
from efilter import query as efilter_query

from plaso.analysis import interface
from plaso.analysis import manager
from plaso.containers import reports


class TaggingAnalysisPlugin(interface.AnalysisPlugin):
  """Analysis plugin that tags events according to rules in a tag file."""

  NAME = u'tagging'

  ENABLE_IN_EXTRACTION = True

  _EVENT_TAG_COMMENT = u'Tag applied by tagging analysis plugin.'

  _OS_TAG_FILES = {
      u'macosx': u'tag_macosx.txt',
      u'windows': u'tag_windows.txt'}

  # A line with no indent is a tag name.
  _TAG_LABEL_LINE = re.compile(r'^(\w+)')
  # A line with leading indent is one of the rules for the preceding tag.
  _TAG_RULE_LINE = re.compile(r'^\s+(.+)')
  # If any of these words are in the query then it's probably objectfilter.
  _OBJECTFILTER_WORDS = re.compile(
      r'\s(is|isnot|equals|notequals|inset|notinset|contains|notcontains)\s')

  def __init__(self):
    """Initializes a tagging analysis plugin."""
    super(TaggingAnalysisPlugin, self).__init__()
    self._autodetect_tag_file_attempt = False
    self._number_of_event_tags = 0
    self._tag_rules = None
    self._tagging_file_name = None

  def _AttemptAutoDetectTagFile(self, analysis_mediator):
    """Detects which tag file is most appropriate.

    Args:
      analysis_mediator (AnalysisMediator): analysis mediator.

    Returns:
      bool: True if a tag file is autodetected.
    """
    self._autodetect_tag_file_attempt = True
    if not analysis_mediator.data_location:
      return False

    platform = analysis_mediator.platform
    filename = self._OS_TAG_FILES.get(platform.lower(), None)
    if not filename:
      return False

    logging.info(u'Using auto detected tag file: {0:s}'.format(filename))
    tag_file_path = os.path.join(analysis_mediator.data_location, filename)
    self.SetAndLoadTagFile(tag_file_path)
    return True

  def _ParseDefinitions(self, tag_file_path):
    """Parses the tag file and yields tuples of label name, list of rule ASTs.

    Args:
      tag_file_path (str): path to the tag file.

    Yields:
      tuple: contains:

        str: label name.
        list[efilter.query.Query]: efilter queries.
    """
    queries = None
    tag = None
    with open(tag_file_path, u'r') as tag_file:
      for line in tag_file.readlines():
        label_match = self._TAG_LABEL_LINE.match(line)
        if label_match:
          if tag and queries:
            yield tag, queries
          queries = []
          tag = label_match.group(1)
          continue

        rule_match = self._TAG_RULE_LINE.match(line)
        if rule_match:
          rule = rule_match.group(1)
          query = self._ParseRule(rule)
          if query:
            queries.append(query)

      # Yield any remaining tags once we reach the end of the file.
      if tag and queries:
        yield tag, queries

  def _ParseRule(self, rule):
    """Parses a single tagging rule.

    This method attempts to detect whether the rule is written with objectfilter
    or dottysql syntax - either is acceptable.

    Example:
      _ParseRule('5 + 5')
      # Returns Sum(Literal(5), Literal(5))

    Args:
      rule (str): rule in either objectfilter or dottysql syntax.

    Returns:
      efilter.query.Query: efilter query of the rule or None.
    """
    if self._OBJECTFILTER_WORDS.search(rule):
      syntax = u'objectfilter'
    else:
      syntax = u'dottysql'

    try:
      return efilter_query.Query(rule, syntax=syntax)

    except efilter_errors.EfilterParseError as exception:
      stripped_rule = rule.rstrip()
      logging.warning(
          u'Unable to build query from rule: "{0:s}" with error: {1:s}'.format(
              stripped_rule, exception.message))

  def _ParseTaggingFile(self, tag_file_path):
    """Parses tag definitions from the source.

    Args:
      tag_file_path (str): path to the tag file.

    Returns:
      efilter.ast.Expression: efilter abstract syntax tree (AST), containing the
          tagging rules.
    """
    tags = []
    for label_name, rules in self._ParseDefinitions(tag_file_path):
      if not rules:
        logging.warning(u'All rules for label "{0:s}" are invalid.'.format(
            label_name))
        continue

      tag = efilter_ast.IfElse(
          # Union will be true if any of the 'rules' match.
          efilter_ast.Union(*[rule.root for rule in rules]),
          # If so then evaluate to a string with the name of the tag.
          efilter_ast.Literal(label_name),
          # Otherwise don't return anything.
          efilter_ast.Literal(None))
      tags.append(tag)

    # Generate a repeated value with all the tags (None will be skipped).
    return efilter_ast.Repeat(*tags)

  def CompileReport(self, mediator):
    """Compiles an analysis report.

    Args:
      mediator (AnalysisMediator): mediates interactions between
          analysis plugins and other components, such as storage and dfvfs.

    Returns:
      AnalysisReport: analysis report.
    """
    report_text = u'Tagging plugin produced {0:d} tags.\n'.format(
        self._number_of_event_tags)
    self._number_of_event_tags = 0
    return reports.AnalysisReport(plugin_name=self.NAME, text=report_text)

  def ExamineEvent(self, mediator, event):
    """Analyzes an EventObject and tags it according to rules in the tag file.

    Args:
      mediator (AnalysisMediator): mediates interactions between analysis
          plugins and other components, such as storage and dfvfs.
      event (EventObject): event to examine.
    """
    if self._tag_rules is None:
      if self._autodetect_tag_file_attempt:
        # There's nothing to tag with, and we've already tried to find a good
        # tag file, so there's nothing we can do with this event (or any other).
        return
      if not self._AttemptAutoDetectTagFile(mediator):
        logging.info(
            u'No tag definition file specified, and plaso was not able to '
            u'autoselect a tagging file. As no definitions were specified, '
            u'no events will be tagged.')
        return

    try:
      matched_labels = efilter_api.apply(self._tag_rules, vars=event)
    except efilter_errors.EfilterTypeError as exception:
      logging.warning(u'Unable to apply efilter query with error: {0:s}'.format(
          exception))
      matched_labels = None

    if not matched_labels:
      return

    labels = list(efilter_api.getvalues(matched_labels))
    event_tag = self._CreateEventTag(event, self._EVENT_TAG_COMMENT, labels)

    mediator.ProduceEventTag(event_tag)
    self._number_of_event_tags += 1

  def SetAndLoadTagFile(self, tagging_file_path):
    """Sets the tag file to be used by the plugin.

    Args:
      tagging_file_path (str): path of the tagging file.
    """
    self._tagging_file_name = tagging_file_path
    self._tag_rules = self._ParseTaggingFile(self._tagging_file_name)


manager.AnalysisPluginManager.RegisterPlugin(TaggingAnalysisPlugin)



# -*- coding:utf-8 -*-
"""Parser for Twitter on iOS 8+ database.

SQLite database path:
/private/var/mobile/Containers/Data/Application/Library/Caches/databases/
SQLite database name: twitter.db
"""

from dfdatetime import posix_time as dfdatetime_posix_time

from plaso.containers import events
from plaso.containers import time_events
from plaso.lib import definitions
from plaso.parsers import sqlite
from plaso.parsers.sqlite_plugins import interface


class TwitterIOSContactEventData(events.EventData):
  """Twitter on iOS 8+ contact event data.

  Attributes:
    description (str): description of the profile.
    followers_count (int): number of accounts following the contact.
    following_count (int): number of accounts the contact is following.
    following (int): 1 if the contact is following the user's account, 0 if not.
    location (str): location of the profile.
    name (str): name of the profile.
    profile_url (str): URL of the profile picture.
    screen_name (str): screen name.
    url (str): URL of the profile.
  """

  DATA_TYPE = u'twitter:ios:contact'

  def __init__(self):
    """Initializes event data."""
    super(TwitterIOSContactEventData, self).__init__(data_type=self.DATA_TYPE)
    self.description = None
    self.followers_count = None
    self.following = None
    self.following_count = None
    self.location = None
    self.name = None
    self.profile_url = None
    self.screen_name = None
    self.url = None


class TwitterIOSStatusEventData(events.EventData):
  """Parent class for Twitter on iOS 8+ status events.

  Attributes:
    favorite_count (int): number of times the status message has been favorited.
    favorited (int): value to mark status as favorite by the account.
    name (str): user's profile name.
    retweet_count (str): number of times the status message has been retweeted.
    text (str): content of the status messsage.
    user_id (int): user unique identifier.
  """

  DATA_TYPE = u'twitter:ios:status'

  def __init__(self):
    """Initializes event data."""
    super(TwitterIOSStatusEventData, self).__init__(data_type=self.DATA_TYPE)
    self.favorite_count = None
    self.favorited = None
    self.name = None
    self.retweet_count = None
    self.text = None
    self.user_id = None


class TwitterIOSPlugin(interface.SQLitePlugin):
  """Parser for Twitter on iOS 8+ database."""

  NAME = u'twitter_ios'
  DESCRIPTION = u'Parser for Twitter on iOS 8+ database'

  QUERIES = [
      ((u'SELECT createdDate, updatedAt, screenName, name, profileImageUrl,'
        u'location, description, url, following, followersCount, followingCount'
        u' FROM Users ORDER BY createdDate'), u'ParseContactRow'),
      ((u'SELECT Statuses.date AS date, Statuses.text AS text, Statuses.userId '
        u'AS user_id, Users.name AS name, Statuses.retweetCount AS '
        u'retweetCount, Statuses.favoriteCount AS favoriteCount, '
        u'Statuses.favorited AS favorited, Statuses.updatedAt AS updatedAt '
        u'FROM Statuses LEFT join Users ON Statuses.userId = Users.id ORDER '
        u'BY date'), u'ParseStatusRow')]

  REQUIRED_TABLES = frozenset([
      u'Lists', u'MyRetweets', u'StatusesShadow', u'UsersShadow',
      u'ListsShadow', u'Statuses', u'Users'])

  SCHEMAS = [{
      u'Lists': (
          u'CREATE TABLE Lists ( \'id\' INTEGER PRIMARY KEY, \'name\' TEXT, '
          u'\'slug\' TEXT, \'desc\' TEXT, \'private\' INTEGER, '
          u'\'subscriberCount\' INTEGER, \'memberCount\' INTEGER, \'userId\' '
          u'INTEGER, \'updatedAt\' REAL )'),
      u'ListsShadow': (
          u'CREATE TABLE ListsShadow ( \'id\' INTEGER PRIMARY KEY, \'name\' '
          u'TEXT, \'slug\' TEXT, \'desc\' TEXT, \'private\' INTEGER, '
          u'\'subscriberCount\' INTEGER, \'memberCount\' INTEGER, \'userId\' '
          u'INTEGER, \'updatedAt\' REAL )'),
      u'MyRetweets': (
          u'CREATE TABLE MyRetweets ( \'statusId\' INTEGER PRIMARY KEY, '
          u'\'myRetweetId\' INTEGER )'),
      u'Statuses': (
          u'CREATE TABLE Statuses ( \'id\' INTEGER PRIMARY KEY, \'text\' TEXT, '
          u'\'date\' REAL, \'userId\' INTEGER, \'inReplyToStatusId\' INTEGER, '
          u'\'retweetedStatusId\' INTEGER, \'geotag\' BLOB, \'entities\' BLOB, '
          u'\'card\' BLOB, \'cardUsers\' BLOB, \'primaryCardType\' INTEGER, '
          u'\'cardVersion\' INTEGER, \'retweetCount\' INTEGER, '
          u'\'favoriteCount\' INTEGER, \'favorited\' INTEGER, \'updatedAt\' '
          u'REAL, \'extraScribeItem\' BLOB, \'withheldScope\' TEXT, '
          u'\'withheldInCountries\' TEXT, \'inReplyToUsername\' TEXT, '
          u'\'possiblySensitive\' INTEGER, \'isPossiblySensitiveAppealable\' '
          u'INTEGER, \'isLifelineAlert\' INTEGER, \'isTruncated\' INTEGER, '
          u'\'previewLength\' INTEGER, \'fullTextLength\' INTEGER, \'lang\' '
          u'TEXT, \'supplmentalLanguage\' TEXT, \'includeInProfileTimeline\' '
          u'INTEGER, \'quotedStatusId\' INTEGER, \'source\' TEXT )'),
      u'StatusesShadow': (
          u'CREATE TABLE StatusesShadow ( \'id\' INTEGER PRIMARY KEY, \'text\' '
          u'TEXT, \'date\' REAL, \'userId\' INTEGER, \'inReplyToStatusId\' '
          u'INTEGER, \'retweetedStatusId\' INTEGER, \'geotag\' BLOB, '
          u'\'entities\' BLOB, \'card\' BLOB, \'cardUsers\' BLOB, '
          u'\'primaryCardType\' INTEGER, \'cardVersion\' INTEGER, '
          u'\'retweetCount\' INTEGER, \'favoriteCount\' INTEGER, \'favorited\' '
          u'INTEGER, \'updatedAt\' REAL, \'extraScribeItem\' BLOB, '
          u'\'withheldScope\' TEXT, \'withheldInCountries\' TEXT, '
          u'\'inReplyToUsername\' TEXT, \'possiblySensitive\' INTEGER, '
          u'\'isPossiblySensitiveAppealable\' INTEGER, \'isLifelineAlert\' '
          u'INTEGER, \'isTruncated\' INTEGER, \'previewLength\' INTEGER, '
          u'\'fullTextLength\' INTEGER, \'lang\' TEXT, '
          u'\'supplementalLanguage\' TEXT, \'includeInProfileTimeline\' '
          u'INTEGER, \'quotedStatusId\' INTEGER, \'source\' TEXT )'),
      u'Users': (
          u'CREATE TABLE Users ( \'id\' INTEGER PRIMARY KEY, \'screenName\' '
          u'TEXT COLLATE NOCASE, \'profileImageUrl\' TEXT, '
          u'\'profileBannerUrl\' TEXT, \'profileLinkColorHexTriplet\' INTEGER, '
          u'\'name\' TEXT, \'location\' TEXT, \'structuredLocation\' BLOB, '
          u'\'description\' TEXT, \'url\' TEXT, \'urlEntities\' BLOB, '
          u'\'bioEntities\' BLOB, \'protected\' INTEGER, \'verified\' INTEGER, '
          u'\'following\' INTEGER, \'deviceFollowing\' INTEGER, '
          u'\'advertiserAccountType\' INTEGER, \'statusesCount\' INTEGER, '
          u'\'mediaCount\' INTEGER, \'favoritesCount\' INTEGER, '
          u'\'followingCount\' INTEGER, \'followersCount\' INTEGER, '
          u'\'followersCountFast\' INTEGER, \'followersCountNormal\' INTEGER, '
          u'\'couldBeStale\' INTEGER, \'isLifelineInstitution\' INTEGER, '
          u'\'hasCollections\' INTEGER, \'updatedAt\' REAL, \'createdDate\' '
          u'REAL, \'isTranslator\' INTEGER, \'hasExtendedProfileFields\' '
          u'INTEGER, \'extendedProfileFields\' BLOB, \'pinnedTweetId\' '
          u'INTEGER, \'businessProfileState\' INTEGER, \'analyticsType\' '
          u'INTEGER )'),
      u'UsersShadow': (
          u'CREATE TABLE UsersShadow ( \'id\' INTEGER PRIMARY KEY, '
          u'\'screenName\' TEXT COLLATE NOCASE, \'profileImageUrl\' TEXT, '
          u'\'profileBannerUrl\' TEXT, \'profileLinkColorHexTriplet\' INTEGER, '
          u'\'name\' TEXT, \'location\' TEXT, \'structuredLocation\' BLOB, '
          u'\'description\' TEXT, \'url\' TEXT, \'urlEntities\' BLOB, '
          u'\'bioEntities\' BLOB, \'protected\' INTEGER, \'verified\' INTEGER, '
          u'\'following\' INTEGER, \'deviceFollowing\' INTEGER, '
          u'\'advertiserAccountType\' INTEGER, \'statusesCount\' INTEGER, '
          u'\'mediaCount\' INTEGER, \'favoritesCount\' INTEGER, '
          u'\'followingCount\' INTEGER, \'followersCount\' INTEGER, '
          u'\'followersCountFast\' INTEGER, \'followersCountNormal\' INTEGER, '
          u'\'couldBeStale\' INTEGER, \'isLifelineInstitution\' INTEGER, '
          u'\'hasCollections\' INTEGER, \'updatedAt\' REAL, \'createdDate\' '
          u'REAL, \'isTranslator\' INTEGER, \'hasExtendedProfileFields\' '
          u'INTEGER, \'extendedProfileFields\' BLOB, \'pinnedTweetId\' '
          u'INTEGER, \'businessProfileState\' INTEGER, \'analyticsType\' '
          u'INTEGER )')}]

  def ParseContactRow(self, parser_mediator, row, query=None, **unused_kwargs):
    """Parses a contact row from the database.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row (sqlite3.Row): row resulting from query.
      query (Optional[str]): query.
    """
    # Note that pysqlite does not accept a Unicode string in row['string'] and
    # will raise "IndexError: Index must be int or string".

    event_data = TwitterIOSContactEventData()
    event_data.description = row['description']
    event_data.followers_count = row['followersCount']
    event_data.following = row['following']
    event_data.following_count = row['followingCount']
    event_data.location = row['location']
    event_data.name = row['name']
    event_data.profile_url = row['profileImageUrl']
    event_data.query = query
    event_data.screen_name = row['screenName']
    event_data.url = row['url']

    timestamp = row['createdDate']
    if timestamp:
      # Convert the floating point value to an integer.
      timestamp = int(timestamp)
      date_time = dfdatetime_posix_time.PosixTime(timestamp=timestamp)
      event = time_events.DateTimeValuesEvent(
          date_time, definitions.TIME_DESCRIPTION_CREATION)
      parser_mediator.ProduceEventWithEventData(event, event_data)

    timestamp = row['updatedAt']
    if timestamp:
      # Convert the floating point value to an integer.
      timestamp = int(timestamp)
      date_time = dfdatetime_posix_time.PosixTime(timestamp=timestamp)
      event = time_events.DateTimeValuesEvent(
          date_time, definitions.TIME_DESCRIPTION_UPDATE)
      parser_mediator.ProduceEventWithEventData(event, event_data)

  def ParseStatusRow(self, parser_mediator, row, query=None, **unused_kwargs):
    """Parses a contact row from the database.

    Args:
      parser_mediator (ParserMediator): mediates interactions between parsers
          and other components, such as storage and dfvfs.
      row (sqlite3.Row): row resulting from query.
      query (Optional[str]): query.
    """
    # Note that pysqlite does not accept a Unicode string in row['string'] and
    # will raise "IndexError: Index must be int or string".

    event_data = TwitterIOSStatusEventData()
    event_data.favorite_count = row['favoriteCount']
    event_data.favorited = row['favorited']
    event_data.name = row['name']
    event_data.query = query
    event_data.retweet_count = row['retweetCount']
    event_data.text = row['text']
    event_data.user_id = row['user_id']

    timestamp = row['date']
    if timestamp:
      # Convert the floating point value to an integer.
      timestamp = int(timestamp)
      date_time = dfdatetime_posix_time.PosixTime(timestamp=timestamp)
      event = time_events.DateTimeValuesEvent(
          date_time, definitions.TIME_DESCRIPTION_CREATION)
      parser_mediator.ProduceEventWithEventData(event, event_data)

    timestamp = row['updatedAt']
    if timestamp:
      # Convert the floating point value to an integer.
      timestamp = int(timestamp)
      date_time = dfdatetime_posix_time.PosixTime(timestamp=timestamp)
      event = time_events.DateTimeValuesEvent(
          date_time, definitions.TIME_DESCRIPTION_UPDATE)
      parser_mediator.ProduceEventWithEventData(event, event_data)


sqlite.SQLiteParser.RegisterPlugin(TwitterIOSPlugin)



# -*- coding: utf-8 -*-
"""The Android contacts2.db database event formatter."""

from plaso.formatters import interface
from plaso.formatters import manager


class AndroidCallFormatter(interface.ConditionalEventFormatter):
  """Formatter for an Android call history event."""

  DATA_TYPE = u'android:event:call'

  FORMAT_STRING_PIECES = [
      u'{call_type}',
      u'Number: {number}',
      u'Name: {name}',
      u'Duration: {duration} seconds']

  FORMAT_STRING_SHORT_PIECES = [u'{call_type} Call']

  SOURCE_LONG = u'Android Call History'
  SOURCE_SHORT = u'LOG'


manager.FormattersManager.RegisterFormatter(AndroidCallFormatter)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""Tests for the Docker JSON parser."""

import unittest

from plaso.lib import timelib
from plaso.parsers import docker

from tests import test_lib as shared_test_lib
from tests.parsers import test_lib


class DockerJSONUnitTest(test_lib.ParserTestCase):
  """Tests for the Docker JSON parser."""

  @shared_test_lib.skipUnlessHasTestFile([
      u'docker', u'containers',
      u'e7d0b7ea5ccf08366e2b0c8afa2318674e8aefe802315378125d2bb83fe3110c',
      u'container-json.log'])
  def testParseContainerLog(self):
    """Tests the _ParseContainerLogJSON function."""
    container_identifier = (
        u'e7d0b7ea5ccf08366e2b0c8afa2318674e8aefe802315378125d2bb83fe3110c')

    parser = docker.DockerJSONParser()
    path_segments = [
        u'docker', u'containers', container_identifier, u'container-json.log']
    storage_writer = self._ParseFile(path_segments, parser)

    self.assertEqual(storage_writer.number_of_events, 10)

    events = list(storage_writer.GetEvents())

    expected_times = [
        u'2016-01-07 16:49:10.000000',
        u'2016-01-07 16:49:10.200000',
        u'2016-01-07 16:49:10.230000',
        u'2016-01-07 16:49:10.237000',
        u'2016-01-07 16:49:10.237200',
        u'2016-01-07 16:49:10.237220',
        u'2016-01-07 16:49:10.237222',
        u'2016-01-07 16:49:10.237222', # losing sub microsec info
        u'2016-01-07 16:49:10.237222',
        u'2016-01-07 16:49:10.237222']

    expected_log = (
        u'\x1b]0;root@e7d0b7ea5ccf: '
        u'/home/plaso\x07root@e7d0b7ea5ccf:/home/plaso# ls\r\n')

    expected_message = (
        u'Text: '
        u'\x1b]0;root@e7d0b7ea5ccf: /home/plaso\x07root@e7d0b7ea5ccf:'
        u'/home/plaso# ls, Container ID: {0:s}, '
        u'Source: stdout').format(container_identifier)
    expected_short_message = (
        u'Text: '
        u'\x1b]0;root@e7d0b7ea5ccf: /home/plaso\x07root@e7d0b7ea5ccf:'
        u'/home/plaso# ls, C...'
        )

    for index, event in enumerate(events):
      expected_timestamp = timelib.Timestamp.CopyFromString(
          expected_times[index])
      self.assertEqual(event.timestamp, expected_timestamp)
      self.assertEqual(event.container_id, container_identifier)
      self.assertEqual(event.log_line, expected_log)
      self.assertEqual(event.log_source, u'stdout')
      self._TestGetMessageStrings(
          event, expected_message, expected_short_message)

  @shared_test_lib.skipUnlessHasTestFile([
      u'docker', u'containers',
      u'e7d0b7ea5ccf08366e2b0c8afa2318674e8aefe802315378125d2bb83fe3110c',
      u'container-json.log'])
  def testParseContainerConfig(self):
    """Tests the _ParseContainerConfigJSON function."""
    container_identifier = (
        u'e7d0b7ea5ccf08366e2b0c8afa2318674e8aefe802315378125d2bb83fe3110c')

    parser = docker.DockerJSONParser()
    path_segments = [
        u'docker', u'containers', container_identifier, u'config.json']
    storage_writer = self._ParseFile(path_segments, parser)

    self.assertEqual(storage_writer.number_of_events, 2)

    events = list(storage_writer.GetEvents())

    event = events[0]
    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2016-01-07 16:49:08.674873')
    self.assertEqual(event.timestamp, expected_timestamp)
    self.assertEqual(event.action, u'Container Started')
    self.assertEqual(event.container_id, container_identifier)
    self.assertEqual(event.container_name, u'e7d0b7ea5ccf')

    event = events[1]
    expected_timestamp = timelib.Timestamp.CopyFromString(
        u'2016-01-07 16:49:08.507979')
    self.assertEqual(event.timestamp, expected_timestamp)
    self.assertEqual(event.action, u'Container Created')
    self.assertEqual(event.container_id, container_identifier)
    self.assertEqual(event.container_name, u'e7d0b7ea5ccf')

  @shared_test_lib.skipUnlessHasTestFile([
      u'docker', u'graph',
      u'3c9a9d7cc6a235eb2de58ca9ef3551c67ae42a991933ba4958d207b29142902b',
      u'json'])
  def testParseLayerConfig(self):
    """Tests the _ParseLayerConfigJSON function."""
    layer_identifier = (
        u'3c9a9d7cc6a235eb2de58ca9ef3551c67ae42a991933ba4958d207b29142902b')

    parser = docker.DockerJSONParser()
    path_segments = [u'docker', u'graph', layer_identifier, u'json']
    storage_writer = self._ParseFile(path_segments, parser)

    self.assertEqual(storage_writer.number_of_events, 1)

    events = list(storage_writer.GetEvents())

    event = events[0]

    expected_command = (
        u'/bin/sh -c sed -i \'s/^#\\s*\\(deb.*universe\\)$/\\1/g\' '
        u'/etc/apt/sources.list')
    self.assertEqual(event.command, expected_command)
    self.assertEqual(event.layer_id, layer_identifier)
    self.assertEqual(event.timestamp, 1444670823079273)
    self.assertEqual(event.timestamp_desc, u'Creation Time')


if __name__ == '__main__':
  unittest.main()



# -*- coding: utf-8 -*-
"""The file system stat event formatter."""

from dfvfs.lib import definitions as dfvfs_definitions

from plaso.formatters import interface
from plaso.formatters import manager
from plaso.lib import errors


class FileStatEventFormatter(interface.ConditionalEventFormatter):
  """The file system stat event formatter."""

  DATA_TYPE = u'fs:stat'

  FORMAT_STRING_PIECES = [
      u'{display_name}',
      u'Type: {file_entry_type}',
      u'({unallocated})']

  FORMAT_STRING_SHORT_PIECES = [
      u'{filename}']

  SOURCE_SHORT = u'FILE'

  _FILE_ENTRY_TYPES = {
      dfvfs_definitions.FILE_ENTRY_TYPE_DEVICE: u'device',
      dfvfs_definitions.FILE_ENTRY_TYPE_DIRECTORY: u'directory',
      dfvfs_definitions.FILE_ENTRY_TYPE_FILE: u'file',
      dfvfs_definitions.FILE_ENTRY_TYPE_LINK: u'link',
      dfvfs_definitions.FILE_ENTRY_TYPE_SOCKET: u'socket',
      dfvfs_definitions.FILE_ENTRY_TYPE_PIPE: u'pipe'}

  def GetMessages(self, unused_formatter_mediator, event):
    """Determines the formatted message strings for an event object.

    Args:
      formatter_mediator (FormatterMediator): mediates the interactions between
          formatters and other components, such as storage and Windows EventLog
          resources.
      event (EventObject): event.

    Returns:
      tuple(str, str): formatted message string and short message string.

    Raises:
      WrongFormatter: if the event object cannot be formatted by the formatter.
    """
    if self.DATA_TYPE != event.data_type:
      raise errors.WrongFormatter(u'Unsupported data type: {0:s}.'.format(
          event.data_type))

    event_values = event.CopyToDict()

    file_entry_type = event_values.get(u'file_entry_type', None)
    if file_entry_type is not None:
      event_values[u'file_entry_type'] = self._FILE_ENTRY_TYPES.get(
          file_entry_type, u'UNKNOWN')

    # The usage of allocated is deprecated in favor of is_allocated but
    # is kept here to be backwards compatible.
    if (not event_values.get(u'allocated', False) and
        not event_values.get(u'is_allocated', False)):
      event_values[u'unallocated'] = u'unallocated'

    return self._ConditionalFormatMessages(event_values)

  def GetSources(self, event):
    """Determines the the short and long source for an event object.

    Args:
      event (EventObject): event.

    Returns:
      tuple(str, str): short and long source string.

    Raises:
      WrongFormatter: if the event object cannot be formatted by the formatter.
    """
    if self.DATA_TYPE != event.data_type:
      raise errors.WrongFormatter(u'Unsupported data type: {0:s}.'.format(
          event.data_type))

    file_system_type = getattr(event, u'file_system_type', u'UNKNOWN')
    timestamp_desc = getattr(event, u'timestamp_desc', u'Time')
    source_long = u'{0:s} {1:s}'.format(file_system_type, timestamp_desc)

    return self.SOURCE_SHORT, source_long


class NTFSFileStatEventFormatter(FileStatEventFormatter):
  """The NTFS file system stat event formatter."""

  DATA_TYPE = u'fs:stat:ntfs'

  FORMAT_STRING_PIECES = [
      u'{display_name}',
      u'File reference: {file_reference}',
      u'Attribute name: {attribute_name}',
      u'Name: {name}',
      u'Parent file reference: {parent_file_reference}',
      u'({unallocated})']

  FORMAT_STRING_SHORT_PIECES = [
      u'{filename}',
      u'{file_reference}',
      u'{attribute_name}']

  SOURCE_SHORT = u'FILE'

  _ATTRIBUTE_NAMES = {
      0x00000010: u'$STANDARD_INFORMATION',
      0x00000030: u'$FILE_NAME'
  }

  def GetMessages(self, unused_formatter_mediator, event):
    """Determines the formatted message strings for an event object.

    Args:
      formatter_mediator (FormatterMediator): mediates the interactions between
          formatters and other components, such as storage and Windows EventLog
          resources.
      event (EventObject): event.

    Returns:
      tuple(str, str): formatted message string and short message string.

    Raises:
      WrongFormatter: if the event object cannot be formatted by the formatter.
    """
    if self.DATA_TYPE != event.data_type:
      raise errors.WrongFormatter(u'Unsupported data type: {0:s}.'.format(
          event.data_type))

    event_values = event.CopyToDict()

    attribute_type = event_values.get(u'attribute_type', 0)
    event_values[u'attribute_name'] = self._ATTRIBUTE_NAMES.get(
        attribute_type, u'UNKNOWN')

    file_reference = event_values.get(u'file_reference', None)
    if file_reference:
      event_values[u'file_reference'] = u'{0:d}-{1:d}'.format(
          file_reference & 0xffffffffffff, file_reference >> 48)

    parent_file_reference = event_values.get(u'parent_file_reference', None)
    if parent_file_reference:
      event_values[u'parent_file_reference'] = u'{0:d}-{1:d}'.format(
          parent_file_reference & 0xffffffffffff, parent_file_reference >> 48)

    if not event_values.get(u'is_allocated', False):
      event_values[u'unallocated'] = u'unallocated'

    return self._ConditionalFormatMessages(event_values)


class NTFSUSNChangeEventFormatter(interface.ConditionalEventFormatter):
  """The NTFS USN change event formatter."""

  DATA_TYPE = u'fs:ntfs:usn_change'

  FORMAT_STRING_PIECES = [
      u'{filename}',
      u'File reference: {file_reference}',
      u'Parent file reference: {parent_file_reference}',
      u'Update source: {update_source}',
      u'Update reason: {update_reason}']

  FORMAT_STRING_SHORT_PIECES = [
      u'{filename}',
      u'{file_reference}',
      u'{update_reason}']

  SOURCE_SHORT = u'FILE'

  _USN_REASON_FLAGS = {
      0x00000001: u'USN_REASON_DATA_OVERWRITE',
      0x00000002: u'USN_REASON_DATA_EXTEND',
      0x00000004: u'USN_REASON_DATA_TRUNCATION',
      0x00000010: u'USN_REASON_NAMED_DATA_OVERWRITE',
      0x00000020: u'USN_REASON_NAMED_DATA_EXTEND',
      0x00000040: u'USN_REASON_NAMED_DATA_TRUNCATION',
      0x00000100: u'USN_REASON_FILE_CREATE',
      0x00000200: u'USN_REASON_FILE_DELETE',
      0x00000400: u'USN_REASON_EA_CHANGE',
      0x00000800: u'USN_REASON_SECURITY_CHANGE',
      0x00001000: u'USN_REASON_RENAME_OLD_NAME',
      0x00002000: u'USN_REASON_RENAME_NEW_NAME',
      0x00004000: u'USN_REASON_INDEXABLE_CHANGE',
      0x00008000: u'USN_REASON_BASIC_INFO_CHANGE',
      0x00010000: u'USN_REASON_HARD_LINK_CHANGE',
      0x00020000: u'USN_REASON_COMPRESSION_CHANGE',
      0x00040000: u'USN_REASON_ENCRYPTION_CHANGE',
      0x00080000: u'USN_REASON_OBJECT_ID_CHANGE',
      0x00100000: u'USN_REASON_REPARSE_POINT_CHANGE',
      0x00200000: u'USN_REASON_STREAM_CHANGE',
      0x00400000: u'USN_REASON_TRANSACTED_CHANGE',
      0x80000000: u'USN_REASON_CLOSE'}

  _USN_SOURCE_FLAGS = {
      0x00000001: u'USN_SOURCE_DATA_MANAGEMENT',
      0x00000002: u'USN_SOURCE_AUXILIARY_DATA',
      0x00000004: u'USN_SOURCE_REPLICATION_MANAGEMENT'}

  def GetMessages(self, unused_formatter_mediator, event):
    """Determines the formatted message strings for an event object.

    Args:
      formatter_mediator (FormatterMediator): mediates the interactions between
          formatters and other components, such as storage and Windows EventLog
          resources.
      event (EventObject): event.

    Returns:
      tuple(str, str): formatted message string and short message string.

    Raises:
      WrongFormatter: if the event object cannot be formatted by the formatter.
    """
    if self.DATA_TYPE != event.data_type:
      raise errors.WrongFormatter(u'Unsupported data type: {0:s}.'.format(
          event.data_type))

    event_values = event.CopyToDict()

    file_reference = event_values.get(u'file_reference', None)
    if file_reference:
      event_values[u'file_reference'] = u'{0:d}-{1:d}'.format(
          file_reference & 0xffffffffffff, file_reference >> 48)

    parent_file_reference = event_values.get(u'parent_file_reference', None)
    if parent_file_reference:
      event_values[u'parent_file_reference'] = u'{0:d}-{1:d}'.format(
          parent_file_reference & 0xffffffffffff, parent_file_reference >> 48)

    update_reason_flags = event_values.get(u'update_reason_flags', 0)
    update_reasons = []
    for bitmask, description in sorted(self._USN_REASON_FLAGS.items()):
      if bitmask & update_reason_flags:
        update_reasons.append(description)

    event_values[u'update_reason'] = u', '.join(update_reasons)

    update_source_flags = event_values.get(u'update_source_flags', 0)
    update_sources = []
    for bitmask, description in sorted(self._USN_SOURCE_FLAGS.items()):
      if bitmask & update_source_flags:
        update_sources.append(description)

    event_values[u'update_source'] = u', '.join(update_sources)

    return self._ConditionalFormatMessages(event_values)


manager.FormattersManager.RegisterFormatters([
    FileStatEventFormatter, NTFSFileStatEventFormatter,
    NTFSUSNChangeEventFormatter])
